#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`3119`-compliant **class-specific utility function** unit
tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.cls.pep.clspep3119` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ raiser : isinstanceable    }....................
def test_die_unless_object_isinstanceable() -> None:
    '''
    Test the
    :func:`beartype._util.cls.pep.clspep3119.die_unless_object_isinstanceable`
    validator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep3119Exception
    from beartype._util.cls.pep.clspep3119 import (
        die_unless_object_isinstanceable)
    from beartype_test.a00_unit.data.data_type import (
        Class,
        NonIsinstanceableClass,
    )
    from pytest import raises

    # ....................{ PASS                           }....................
    # Assert this raiser accepts an isinstanceable type.
    die_unless_object_isinstanceable(Class)

    # Assert this raiser accepts a tuple of isinstanceable types.
    die_unless_object_isinstanceable((Class, str))

    # Assert this raiser accepts a PEP 604-compliant new-style union of
    # isinstanceable types.
    die_unless_object_isinstanceable(Class | str)

    # ....................{ FAIL                           }....................
    # Assert this raiser raises the expected exception when passed a
    # non-isinstanceable object.
    with raises(BeartypeDecorHintPep3119Exception):
        die_unless_object_isinstanceable('Moab.')

    # Assert this raiser raises the expected exception when passed a
    # non-isinstanceable type.
    with raises(BeartypeDecorHintPep3119Exception):
        die_unless_object_isinstanceable(NonIsinstanceableClass)

    # Assert this raiser raises the expected exception when passed a tuple of
    # one isinstanceable type and one non-isinstanceable type.
    with raises(BeartypeDecorHintPep3119Exception):
        die_unless_object_isinstanceable((Class, NonIsinstanceableClass))

    # Assert this raiser raises the expected exception when passed a PEP
    # 604-compliant new-style union of one isinstanceable type and one
    # non-isinstanceable type.
    with raises(BeartypeDecorHintPep3119Exception):
        die_unless_object_isinstanceable(Class | NonIsinstanceableClass)


def test_die_unless_type_isinstanceable() -> None:
    '''
    Test the
    :func:`beartype._util.cls.pep.clspep3119.die_unless_type_isinstanceable`
    validator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep3119Exception
    from beartype._util.cls.pep.clspep3119 import (
        die_unless_type_isinstanceable)
    from beartype_test.a00_unit.data.data_type import (
        Class,
        NonIsinstanceableClass,
    )
    from pytest import raises

    # ....................{ PASS                           }....................
    # Assert this validator accepts an isinstanceable type.
    die_unless_type_isinstanceable(Class)

    # ....................{ FAIL                           }....................
    # Assert this validator raises the expected exception when passed a
    # non-isinstanceable type.
    with raises(BeartypeDecorHintPep3119Exception):
        die_unless_type_isinstanceable(NonIsinstanceableClass)

    # Assert this validator raises the expected exception when passed a
    # non-type.
    with raises(BeartypeDecorHintPep3119Exception):
        die_unless_type_isinstanceable('Moab.')

# ....................{ TESTS ~ raiser : issubclassable    }....................
def test_die_unless_object_issubclassable() -> None:
    '''
    Test the
    :func:`beartype._util.cls.pep.clspep3119.die_unless_object_issubclassable`
    validator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep3119Exception
    from beartype._util.cls.pep.clspep3119 import (
        die_unless_object_issubclassable)
    from beartype_test.a00_unit.data.data_type import (
        Class,
        NonIssubclassableClass,
    )
    from pytest import raises

    # ....................{ PASS                           }....................
    # Assert this raiser accepts an issubclassable type.
    die_unless_object_issubclassable(Class)

    # Assert this raiser accepts a tuple of issubclassable types.
    die_unless_object_issubclassable((Class, str))

    # Assert this raiser accepts a PEP 604-compliant new-style union of
    # issubclassable types.
    die_unless_object_issubclassable(Class | str)

    # ....................{ FAIL                           }....................
    # Assert this raiser raises the expected exception when passed a
    # non-issubclassable object.
    with raises(BeartypeDecorHintPep3119Exception):
        die_unless_object_issubclassable('Moab.')

    # Assert this raiser raises the expected exception when passed a
    # non-issubclassable type.
    with raises(BeartypeDecorHintPep3119Exception):
        die_unless_object_issubclassable(NonIssubclassableClass)

    # Assert this raiser raises the expected exception when passed a tuple of
    # one issubclassable type and one non-issubclassable type.
    with raises(BeartypeDecorHintPep3119Exception):
        die_unless_object_issubclassable((Class, NonIssubclassableClass))

    # Assert this raiser raises the expected exception when passed a PEP
    # 604-compliant new-style union of one issubclassable type and one
    # non-issubclassable type.
    with raises(BeartypeDecorHintPep3119Exception):
        die_unless_object_issubclassable(Class | NonIssubclassableClass)


def test_die_unless_type_issubclassable() -> None:
    '''
    Test the
    :func:`beartype._util.cls.pep.clspep3119.die_unless_type_issubclassable`
    validator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep3119Exception
    from beartype._util.cls.pep.clspep3119 import (
        die_unless_type_issubclassable)
    from beartype_test.a00_unit.data.data_type import (
        Class,
        NonIssubclassableClass,
    )
    from pytest import raises

    # ....................{ PASS                           }....................
    # Assert this validator accepts an issubclassable type.
    die_unless_type_issubclassable(Class)

    # ....................{ FAIL                           }....................
    # Assert this validator raises the expected exception when passed a
    # non-issubclassable type.
    with raises(BeartypeDecorHintPep3119Exception):
        die_unless_type_issubclassable(NonIssubclassableClass)

    # Assert this validator raises the expected exception when passed a
    # non-type.
    with raises(BeartypeDecorHintPep3119Exception):
        die_unless_type_issubclassable('Moab.')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`586`-compliant type hint utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep586Exception
from beartype._data.cls.datacls import TYPES_PEP586_ARG
from beartype._data.typing.datatypingport import Hint
from beartype._data.typing.datatyping import TypeException
from beartype._data.hint.sign.datahintsigns import HintSignLiteral
from beartype._util.text.utiltextjoin import join_delimited_disjunction_types

# ....................{ VALIDATORS                         }....................
def die_unless_hint_pep586(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep586Exception,
    exception_prefix: str = '',
) -> None:
    '''
    Raise an exception of the passed type unless the passed object is a
    :pep:`586`-compliant type hint (i.e., subscription of either the
    :attr:`typing.Literal` or :attr:`typing_extensions.Literal` type hint
    factories).

    Ideally, the :attr:`typing.Literal` singleton would internally validate the
    literal objects subscripting that singleton at subscription time (i.e., in
    the body of the ``__class_getitem__()`` dunder method). Whereas *all* other
    :mod:`typing` attributes do just that, :attr:`typing.Literal` permissively
    accepts all possible arguments like a post-modern philosopher hopped up on
    too much tenure. For inexplicable reasons, :pep:`586` explicitly requires
    third-party type checkers (that's us) to validate these hints rather than
    standardizing that validation in the :mod:`typing` module. Weep, Guido!

    Caveats
    -------
    **This function is slow** and should thus be called only once per
    visitation of a :pep:`586`-compliant type hint. Specifically, this function
    is :math:`O(n)` for :math:`n` the number of arguments subscripting this
    hint.

    Parameters
    ----------
    hint : Hint
        Object to be inspected.
    exception_cls : TypeException, optional
        Type of exception to be raised in the event of fatal error. Defaults to
        :exc:`.BeartypeDecorHintPep586Exception`.
    exception_prefix : str, optional
        Human-readable substring prefixing the representation of this object in
        the exception message. Defaults to the empty string.

    Raises
    ------
    exception_cls
        If this object either:

        * Is *not* a subscription of either the :attr:`typing.Literal` or
          :attr:`typing_extensions.Literal` type hint factories.
        * Subscripts either factory with zero arguments via the empty tuple,
          which these factories sadly fails to guard against.
        * Subscripts either factory with one or more arguments that are *not*
          **valid literals**, defined as the set of all:

          * Booleans.
          * Byte strings.
          * Integers.
          * Unicode strings.
          * :class:`enum.Enum` members.
          * The :data:`None` singleton.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepget import get_hint_pep_args
    from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign

    # If this hint is *NOT* PEP 586-compliant, raise an exception.
    if get_hint_pep_sign(hint) is not HintSignLiteral:
        raise exception_cls(
            f'{exception_prefix}type hint {repr(hint)} not PEP 586-compliant '
            f'(e.g., "typing.Literal[...]", "typing_extensions.Literal[...]").'
        )
    # Else, this hint is PEP 586-compliant.

    # Tuple of zero or more literal objects subscripting this hint.
    hint_literals = get_hint_pep_args(hint)

    # If this hint is unsubscripted...
    if not hint_literals:
        # Exception message to be raised.
        exception_message = f'{exception_prefix}PEP 586 type hint {repr(hint)} '

        # If this hint defines the standard "__args__" dunder attribute, this
        # hint *MUST* have been subscripted by the the empty tuple. Ideally, the
        # "typing.Literal" factory would guard against this itself. It does not;
        # thus, we do. Construct an appropriate message.
        if hasattr(hint_literals, '__args__'):
            exception_message += (
                'subscripted by empty tuple, '
                'which is not a valid literal object.'
            )
        # Else, this hint fails to define the standard "__args__" dunder
        # attribute. In this case, this hint *MUST* be the unsubscripted
        # "typing.Literal" factory -- which conveys *NO* meaningful semantics
        # and is thus invalid as a type hint. Construct an appropriate message.
        else:
            exception_message += (
                'unsubscripted (i.e., subscripted by no literal objects).')

        # Raise this exception.
        raise exception_cls(exception_message)
    # If this hint is subscripted by one or more literal objects.

    # For each argument subscripting this hint...
    #
    # Sadly, despite PEP 586 imposing strict restrictions on the types of
    # objects permissible as arguments subscripting the "typing.Literal"
    # singleton, PEP 586 explicitly offloads the odious chore of enforcing those
    # restrictions onto third-party type checkers by intentionally implementing
    # that singleton to permissively accept *ALL* possible objects when
    # subscripted:
    #     Although the set of parameters Literal[...] may contain at type check
    #     time is very small, the actual implementation of typing.Literal will
    #     not perform any checks at runtime.
    for hint_literal_index, hint_literal in enumerate(hint_literals):
        # If this argument is invalid as a literal argument...
        if not isinstance(hint_literal, TYPES_PEP586_ARG):
            # Human-readable concatenation of the types of all valid literal
            # arguments, delimited by commas and/or "or".
            hint_literal_types = join_delimited_disjunction_types(
                TYPES_PEP586_ARG)

            # Raise an exception.
            raise exception_cls(
                f'{exception_prefix}PEP 586 type hint {repr(hint)} '
                f'argument {hint_literal_index} '
                f'{repr(hint_literal)} not {hint_literal_types}.'
            )
        # Else, this argument is valid as a literal argument.

# ....................{ GETTERS                            }....................
#FIXME: Unit test us up, please.
def get_hint_pep586_literals(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep586Exception,
    exception_prefix: str = '',
) -> tuple:
    '''
    Tuple of zero or more literal objects subscripting the passed
    :pep:`586`-compliant type hint (i.e., subscription of either the
    :attr:`typing.Literal` or :attr:`typing_extensions.Literal` type hint
    factories).

    This getter is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Caveats
    -------
    **This low-level getter performs no validation of the contents of this
    tuple.** Consider calling the high-level :func:`die_unless_hint_pep586`
    validator to do so before leveraging this tuple elsewhere.

    Parameters
    ----------
    hint : Hint
        :pep:`586`-compliant type hint to be inspected.
    exception_cls : TypeException
        Type of exception to be raised. Defaults to
        :exc:`BeartypeDecorHintPep586Exception`.
    exception_prefix : str, optional
        Human-readable substring prefixing the representation of this object in
        the exception message. Defaults to the empty string.

    Returns
    -------
    tuple
        Tuple of zero or more literal objects subscripting this hint.

    Raises
    ------
    exception_cls
        If this object is *not* a :pep:`586`-compliant type hint.
    '''

    # If this hint is *NOT* PEP 586-compliant, raise an exception.
    die_unless_hint_pep586(hint=hint, exception_prefix=exception_prefix)
    # Else, this hint is PEP 586-compliant.

    # Return the standard tuple of all literals subscripting this hint.
    return hint.__args__  # pyright: ignore



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **path copiers** (i.e., low-level callables permanently copying
on-disk files and directories in various reasonably safe and portable ways).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilPathDirException
from beartype.typing import Optional
from beartype._data.typing.datatyping import (
    CollectionStrs,
    PathnameLike,
    PathnameLikeTuple,
    TypeException,
)
from collections.abc import Callable
from pathlib import Path
from enum import (
    Enum,
    auto as next_enum_member_value,
    unique as die_unless_enum_member_values_unique,
)
from shutil import (
    copytree,
    ignore_patterns,
)

# ....................{ ENUMERATIONS                       }....................
@die_unless_enum_member_values_unique
class BeartypeDirCopyOverwritePolicy(Enum):
    '''
    Enumeration of all kinds of **directory-copying overwrite policies** (i.e.,
    competing strategies for handling edge cases in which a target path already
    exists when recursively copying directories, each with concomitant tradeoffs
    with respect to safety).

    Note that enumeration members are intentionally ordered from most safe
    (:attr:`.HALT_WITH_EXCEPTION`) to least safe (:attr:`.OVERWRITE`).

    Attributes
    ----------
    HALT_WITH_EXCEPTION : EnumMemberType
        Policy raising a fatal exception if any target path already exists. This
        constitutes the strictest and thus safest such policy.
    SKIP_WITH_WARNING : EnumMemberType
        Policy ignoring (i.e., skipping) each existing target path with a
        non-fatal warning. This policy strikes a comfortable balance between
        strictness and laxness and is thus the recommended default.
    OVERWRITE : EnumMemberType
        Policy silently overwriting each existing target path. This constitutes
        the laxest and thus riskiest such policy.
    '''

    HALT_WITH_EXCEPTION = next_enum_member_value()
    SKIP_WITH_WARNING = next_enum_member_value()
    OVERWRITE = next_enum_member_value()

# ....................{ COPIERS                            }....................
#FIXME: Unit test us up, please.
def copy_dir(
    # Mandatory parameters.
    src_dirname: PathnameLike,
    trg_dirname: PathnameLike,

    # Optional parameters.
    overwrite_policy: BeartypeDirCopyOverwritePolicy = (
        BeartypeDirCopyOverwritePolicy.HALT_WITH_EXCEPTION),
    ignore_basename_globs: Optional[CollectionStrs] = None,
    exception_cls: TypeException = _BeartypeUtilPathDirException,
    exception_prefix: str = '',
) -> None:
    '''
    Recursively copy the source directory with the passed dirname to the
    target directory with the passed dirname.

    For generality:

    * All nonexistent parents of the target directory will be recursively
      created, mimicking the action of the ``mkdir -p`` shell command on
      POSIX-compatible platforms in a platform-agnostic manner.
    * All symbolic links in the source directory will be preserved (i.e.,
      copied as is rather than their transitive targets copied instead).

    Caveats
    -------
    **This function is subject to subtle race conditions if multiple threads
    and/or processes concurrently attempt to mutate any relevant path on the
    local filesystem.** Since *all* filesystem-centric logic suffers similar
    issues, we leave this issue as an exercise for the caller.

    Parameters
    ----------
    src_dirname : PathnameLike
        Absolute or relative dirname of the source directory to be copied from.
    trg_dirname : PathnameLike
        Absolute or relative dirname of the target directory to be copied to.
    overwrite_policy : BeartypeDirCopyOverwritePolicy, default: BeartypeDirCopyOverwritePolicy.HALT_WITH_EXCEPTION
        **Directory overwrite policy** (i.e., strategy for handling existing
        paths to be overwritten by this copy). Defaults to
        :attr:`BeartypeDirCopyOverwritePolicy.HALT_WITH_EXCEPTION`, raising an
        exception if any target path already exists.
    ignore_basename_globs : Collection[str] | None, default: None
        Collection of shell-style globs (e.g., ``('*.tmp', '.keep')``) matching
        the basenames of all paths transitively owned by this source directory
        to be ignored during recursion and hence neither copied nor visited.
        Defaults to ``None``, in which case *all* paths transitively owned by
        this source directory are unconditionally copied and visited.

        Note this parameter is incompatible with the
        :attr:`BeartypeDirCopyOverwritePolicy.OVERWRITE` policy. If this
        parameter is non-:data:`None` and the ``overwrite_policy`` parameter is
        :attr:`BeartypeDirCopyOverwritePolicy.OVERWRITE`, an exception is
        raised.
    exception_cls : Type[Exception], default: _BeartypeUtilPathException
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`._BeartypeUtilPathException`.
    exception_prefix : str, default: ''
        Human-readable substring prefixed raised exceptions messages. Defaults
        to the empty string.

    Raises
    ------
    exception_cls
        If either:

        * The source directory does *not* exist.
        * The target directory is a subdirectory of the source directory.
          Permitting this edge case induces non-trivial issues, including
          infinite recursion from within the musty entrails of the
          :mod:`distutils` package (e.g., due to relative symbolic links).
        * The passed ``overwrite_policy`` parameter is
          :attr:`BeartypeDirCopyOverwritePolicy.HALT_WITH_EXCEPTION` *and* one or more
          subdirectories of the target directory already exist that are also
          subdirectories of the source directory. For safety, this function
          always preserves rather than overwrites existing target
          subdirectories.

    See Also
    -----------
    https://stackoverflow.com/a/22588775/2809027
        StackOverflow answer strongly inspiring this function's
        :attr:`BeartypeDirCopyOverwritePolicy.SKIP_WITH_WARNING` implementation.
    '''
    assert isinstance(src_dirname, PathnameLikeTuple), (
        f'{repr(src_dirname)} neither string nor "Path" object.')
    assert isinstance(trg_dirname, PathnameLikeTuple), (
        f'{repr(trg_dirname)} neither string nor "Path" object.')

    # ....................{ IMPORTS                        }....................
    # Avoid circular import dependencies.
    from beartype._util.path.utilpathtest import (
        die_if_dir,
        die_if_subpath,
        die_unless_dir,
    )

    # ....................{ PREAMBLE                       }....................
    # High-level "Path" objects encapsulating these dirnames.
    src_dirname = Path(src_dirname)
    trg_dirname = Path(trg_dirname)

    # If the source directory does *NOT* exist, raise an exception.
    die_unless_dir(
        dirname=src_dirname,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )

    # If the target directory is a subdirectory of the source directory, raise
    # an exception. Permitting this edge case provokes issues, including
    # infinite recursion from within the musty entrails of the "distutils"
    # codebase (possibly due to relative symbolic links).
    die_if_subpath(
        parent_pathname=src_dirname,
        child_pathname=trg_dirname,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )

    # If passed an iterable of shell-style globs matching ignorable basenames,
    # convert this iterable into a predicate function of the form required by
    # the shutil.copytree() function. Specifically, this function accepts the
    # absolute or relative pathname of an arbitrary directory and an iterable
    # of the basenames of all subdirectories and files directly in this
    # directory; this function returns an iterable of the basenames of all
    # subdirectories and files in this directory to be ignored. This signature
    # resembles:
    #
    #     def ignore_basename_func(
    #         parent_dirname: str,
    #         child_basenames: IterableTypes) -> IterableTypes
    ignore_basename_func: Optional[Callable] = None
    if ignore_basename_globs is not None:
        ignore_basename_func = ignore_patterns(*ignore_basename_globs)

    # ....................{ POLICIES                       }....................
    # If either:
    # * Raising a fatal exception if any target path already exists *OR*...
    # * Overwriting this target directory with this source directory...
    #
    # Then the standard shutil.copytree() function applies to this use case.
    if overwrite_policy in _COPY_DIR_OVERWRITE_POLICIES_COPYTREE:
        # Dictionary of all keyword arguments to pass to shutil.copytree(),
        # preserving symbolic links as is.
        copytree_kwargs: dict = {
            'symlinks': True,
        }

        # If raising a fatal exception if any target path already exists, do so.
        # While we could defer to the exception raised by the shutil.copytree()
        # function for this case, this exception's message erroneously refers to
        # this directory as a file and is hence best avoided as unreadable:
        #     [Errno 17] File exists: 'sample_sim'  # <-- lolbro! useless.
        if (
            overwrite_policy is
            BeartypeDirCopyOverwritePolicy.HALT_WITH_EXCEPTION
        ):
            die_if_dir(
                dirname=trg_dirname,
                exception_cls=exception_cls,
                exception_prefix=exception_prefix,
            )
        # Else, this target directory is being overwritten by this source
        # directory. In this case, silently accept this target directory if this
        # directory already exists.
        else:
            copytree_kwargs['dirs_exist_ok'] = True

        # If ignoring basenames, inform shutil.copytree() of these basenames.
        if ignore_basename_func is not None:
            copytree_kwargs['ignore'] = ignore_basename_func
        # Else, no basenames are being ignored.

        # Recursively copy this source to target directory. To avoid silently
        # overwriting all conflicting target paths, the shutil.copytree()
        # rather than dir_util.copy_tree() function is called.
        copytree(src=src_dirname, dst=trg_dirname, **copytree_kwargs)

    #FIXME: Given how awesomely flexible the manual approach implemented below
    #is, we should probably consider simply rewriting the above two approaches
    #to reuse the exact same logic. It works. It's preferable. Let's reuse it.
    #FIXME: Actually, this is increasingly critical. Third-party functions
    #called above -- notably, the dir_util.copy_tree() function -- appear to
    #suffer critical edge cases. This can be demonstrated via the BETSEE GUI by
    #attempting to save an opened simulation configuration to a subdirectory of
    #itself, which appears to provoke infinite recursion from within the musty
    #depths of the "distutils" codebase. Of course, the implementation below
    #could conceivably suffer similar issues. If this is the case, this
    #function should explicitly detect attempts to recursively copy a source
    #directory into a subdirectory of itself and raise an exception.
    #FIXME: Uncomment this if and when we actually need it. So, probably never.
    #Refactoring this out of the BETSE codebase is tedious beyond belief. *sigh*
    # # Else if logging a warning for each target path that already exists, do so
    # # by manually implementing recursive directory copying. Sadly, Python
    # # provides no means of doing so "out of the box."
    # elif overwrite_policy is BeartypeDirCopyOverwritePolicy.SKIP_WITH_WARNING:
    #     # Avoid circular import dependencies.
    #     from betse.util.path import files, paths, pathnames
    #     from betse.util.type.iterable import sequences
    #
    #     # Passed parameters renamed for disambiguity.
    #     src_root_dirname = src_dirname
    #     trg_root_dirname = trg_dirname
    #
    #     # Basename of the top-level target directory to be copied to.
    #     trg_root_basename = pathnames.get_basename(src_root_dirname)
    #
    #     # For the absolute pathname of each recursively visited source
    #     # directory, an iterable of the basenames of all subdirectories of this
    #     # directory, and an iterable of the basenames of all files of this
    #     # directory...
    #     for src_parent_dirname, subdir_basenames, file_basenames in _walk(
    #         src_root_dirname):
    #         # Relative pathname of the currently visited source directory
    #         # relative to the absolute pathname of this directory.
    #         parent_dirname_relative = pathnames.relativize(
    #             src_dirname=src_root_dirname, trg_pathname=src_parent_dirname)
    #
    #         # If ignoring basenames...
    #         if ignore_basename_func is not None:
    #             # Sets of the basenames of all ignorable subdirectories and
    #             # files of this source directory.
    #             subdir_basenames_ignored = ignore_basename_func(
    #                 src_parent_dirname, subdir_basenames)
    #             file_basenames_ignored = ignore_basename_func(
    #                 src_parent_dirname, file_basenames)
    #
    #             # If ignoring one or more subdirectories...
    #             if subdir_basenames_ignored:
    #                 # Log the basenames of these subdirectories.
    #                 logs.log_debug(
    #                     'Ignoring source "%s/%s" subdirectories: %r',
    #                     trg_root_basename,
    #                     parent_dirname_relative,
    #                     subdir_basenames_ignored)
    #
    #                 # Remove these subdirectories from the original iterable.
    #                 # Since the os.walk() function supports in-place changes to
    #                 # this iterable, this iterable is modified via this less
    #                 # efficient function rather than efficient alternatives
    #                 # (e.g., set subtraction).
    #                 sequences.remove_items(
    #                     sequence=subdir_basenames,
    #                     items=subdir_basenames_ignored)
    #
    #             # If ignoring one or more files...
    #             if file_basenames_ignored:
    #                 # Log the basenames of these files.
    #                 logs.log_debug(
    #                     'Ignoring source "%s/%s" files: %r',
    #                     trg_root_basename,
    #                     parent_dirname_relative,
    #                     file_basenames_ignored)
    #
    #                 # Remove these files from the original iterable. Unlike
    #                 # above, we could technically modify this iterable via
    #                 # set subtraction: e.g.,
    #                 #
    #                 #     subdir_basenames -= subdir_basenames_ignored
    #                 #
    #                 # For orthogonality, preserve the above approach instead.
    #                 sequences.remove_items(
    #                     sequence=file_basenames,
    #                     items=file_basenames_ignored)
    #
    #         # Absolute pathname of the corresponding target directory.
    #         trg_parent_dirname = pathnames.join(
    #             trg_root_dirname, parent_dirname_relative)
    #
    #         # Create this target directory if needed.
    #         make_unless_dir(trg_parent_dirname)
    #
    #         # For the basename of each non-ignorable file of this source
    #         # directory...
    #         for file_basename in file_basenames:
    #             # Absolute filenames of this source and target file.
    #             src_filename = pathnames.join(
    #                 src_parent_dirname, file_basename)
    #             trg_filename = pathnames.join(
    #                 trg_parent_dirname, file_basename)
    #
    #             # If this target file already exists...
    #             if paths.is_path(trg_filename):
    #                 # Relative filename of this file. The absolute filename of
    #                 # this source or target file could be logged instead, but
    #                 # this relative filename is significantly more terse.
    #                 filename_relative = pathnames.join(
    #                     trg_root_basename,
    #                     parent_dirname_relative,
    #                     file_basename)
    #
    #                 # Warn of this file being ignored.
    #                 logs.log_warning(
    #                     'Ignoring existing target file: %s', filename_relative)
    #
    #                 # Ignore this file by continuing to the next.
    #                 continue
    #
    #             # Copy this source to target file.
    #             files.copy(
    #                 src_filename=src_filename, trg_filename=trg_filename)
    # Else, this overwrite policy is unrecognized. Raise an exception.
    else:
        raise exception_cls(
            f'{exception_prefix}'
            f'overwrite policy "{overwrite_policy}" unrecognized.'
        )

# ....................{ PRIVATE ~ constants                }....................
_COPY_DIR_OVERWRITE_POLICIES_COPYTREE = frozenset((
    BeartypeDirCopyOverwritePolicy.HALT_WITH_EXCEPTION,
    BeartypeDirCopyOverwritePolicy.OVERWRITE,
))
'''
Frozen set of all **copytree-friendly directory overwrite policies** (i.e.,
:class:`.BeartypeDirCopyOverwritePolicy` enumeration members suitable for
passing as the ``overwrite_policy`` parameter to the :func:`.copy_dir` function
such that the resulting implementation reduces to a trivial call of the standard
:func:`shutil.copytree` function).
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype** :pep:`484`-compliant **union type hint violation describers**
(i.e., functions returning human-readable strings explaining violations of
:pep:`484`-compliant union type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeCallHintPepRaiseException
from beartype._data.hint.sign.datahintsignset import HINT_SIGNS_UNION
from beartype._check.error.errcause import ViolationCause
from beartype._check.metadata.hint.hintsane import HINT_SANE_IGNORABLE
from beartype._util.hint.pep.utilpepget import (
    get_hint_pep_origin_type_isinstanceable_or_none)
from beartype._util.hint.pep.utilpeptest import is_hint_pep
from beartype._util.text.utiltextjoin import join_delimited_disjunction_types
from beartype._util.text.utiltextmunge import (
    suffix_str_unless_suffixed,
    uppercase_str_char_first,
)
from beartype._util.text.utiltextrepr import represent_pith

# ....................{ GETTERS                            }....................
def find_cause_pep484604_union(cause: ViolationCause) -> ViolationCause:
    '''
    Output cause describing whether the pith of the passed input cause either
    satisfies or violates the PEP-compliant union type hint of that cause.

    Parameters
    ----------
    cause : ViolationCause
        Input cause providing this data.

    Returns
    -------
    ViolationCause
        Output cause type-checking this data.
    '''
    assert isinstance(cause, ViolationCause), f'{repr(cause)} not cause.'
    assert cause.hint_sign in HINT_SIGNS_UNION, (
        f'{repr(cause.hint)} not union sign.')
    # print(f'[union] Finding cause for child hints {cause.hint_childs_sane}...')

    # ....................{ LOCALS                         }....................
    # Indentation preceding each line of the strings returned by child getter
    # functions called by this parent getter function, offset to visually
    # demarcate child from parent causes in multiline strings.
    CAUSE_INDENT_CHILD = cause.cause_indent + '  '

    # List of all human-readable strings describing the failure of this pith to
    # satisfy each of these child hints.
    cause_strs = []

    # Subset of all classes shallowly associated with these child hints (i.e.,
    # by being either these child hints in the case of non-"typing" classes
    # *OR* the classes originating these child hints in the case of
    # PEP-compliant type hints) that this pith fails to shallowly satisfy.
    hint_types_violated = set()

    # Truncated object representation of this pith.
    pith_repr = represent_pith(cause.pith)

    # 0-based index of the first non-whitespace character following this
    # representation in violation causes collected below. Look. Just accept it.
    PITH_REPR_INDEX = len(pith_repr) + 1

    # ....................{ SEARCH                         }....................
    # For each subscripted argument of this union...
    for hint_child_sane in cause.hint_childs_sane:
        # If this child hint is ignorable, continue to the next.
        if hint_child_sane is HINT_SANE_IGNORABLE:
            continue
        # Else, this child hint is unignorable.

        # Child hint encapsulated by this metadata.
        hint_child = hint_child_sane.hint

        # If this child hint is PEP-compliant...
        if is_hint_pep(hint_child):
            # Non-"typing" class originating this child hint if any *OR* "None"
            # otherwise.
            hint_child_origin_type = (
                get_hint_pep_origin_type_isinstanceable_or_none(hint_child))

            # If...
            if (
                # This child hint originates from a non-"typing" type *AND*...
                hint_child_origin_type is not None and
                # This pith is *NOT* an instance of this type...
                not isinstance(cause.pith, hint_child_origin_type)
            # Then this pith fails to satisfy this child hint. In this case...
            ):
                # Add this type to the subset of all types this pith violates.
                hint_types_violated.add(hint_child_origin_type)

                # Continue to the next child hint.
                continue
            # Else, this pith is an instance of this type and thus shallowly
            # (but *NOT* necessarily deeply) satisfies this child hint.

            # Child hint output cause to be returned, type-checking only whether
            # this pith deeply satisfies this child hint.
            cause_child = cause.permute_cause(
                hint_sane=hint_child_sane, cause_indent=CAUSE_INDENT_CHILD,
            ).find_cause()

            # If this pith deeply satisfies this child hint, return this cause.
            if cause_child.cause_str_or_none is None:
                # print('Union child {!r} pith {!r} deeply satisfied!'.format(hint_child, pith))
                return cause
            # Else, this pith deeply violates this child hint.

            # Cause of this violation.
            cause_str = cause_child.cause_str_or_none

            # If this cause is prefixed by the truncated object representation
            # of this pith...
            #
            # Note that this should *ALWAYS* be the case. Nonetheless, let's
            # *NOT* assume anything to avoid exploding everything.
            if cause_str.startswith(pith_repr):
                # Strip the prefixing
                # representation of this pith from this cause (e.g., the prefix
                # "MuhClass <object MuhClass at 0x7fbc277a2cf0>" from the cause
                # 'MuhClass <object MuhClass at 0x7fbc277a2cf0> not instance of
                # <protocol "muh_package.MuhProtocol">'). Why? Because the block
                # of text preceding the bulleted list containing this cause is
                # already redundantly prefixed by this representation.
                cause_str = cause_str[PITH_REPR_INDEX:]
            # Else, this cause is *NOT* prefixed by the truncated object
            # representation of this pith. In this case, silently accept that
            # Bad Things have happened and that we should move to a Bad Future.

            # Append the cause of this violation as a bullet-prefixed line to
            # the running list of these lines.
            cause_strs.append(cause_str)
            # print(f'[union] Appended PEP-compliant child hint {hint_child} cause {cause_str}!')
        # Else, this child hint is PEP-noncompliant. In this case...
        else:
            # Assert this child hint to be a non-"typing" class. Note that
            # the "typing" module should have already guaranteed that all
            # subscripted arguments of unions are either PEP-compliant type
            # hints or non-"typing" classes.
            assert isinstance(hint_child, type), (
                f'{cause.exception_prefix}union type hint '
                f'{repr(cause.hint)} child hint {repr(hint_child)} invalid '
                f'(i.e., neither type hint nor non-"typing" class).')
            # Else, this child hint is a non-"typing" type.

            # If this pith is an instance of this type, this pith satisfies this
            # hint. In this case, return this cause as is.
            if isinstance(cause.pith, hint_child):
                return cause
            # Else, this pith is *NOT* an instance of this type, implying this
            # pith to *NOT* satisfy this hint.

            # Add this class to the subset of all types this pith violates.
            hint_types_violated.add(hint_child)
            # print(f'[union] Appended PEP-noncompliant child hint {hint_child}!')

    # ....................{ CAUSE                          }....................
    # If this pith fails to shallowly satisfy one or more of the types of this
    # union, concatenate these failures onto one discrete bullet-prefixed line.
    if hint_types_violated:
        # Human-readable comma-delimited disjunction of the names of these
        # classes (e.g., "bool, float, int, or str").
        cause_types_unsatisfied = join_delimited_disjunction_types(
            types=hint_types_violated, is_color=cause.conf.is_color)

        # Prepend this cause as a discrete bullet-prefixed line.
        #
        # Note that this cause is intentionally prependend rather than appended
        # to this list. Since this cause applies *ONLY* to the shallow type of
        # the current pith rather than any items contained in this pith,
        # listing this shallow cause *BEFORE* other deeper causes typically
        # applying to items contained in this pith produces substantially more
        # human-readable exception messages: e.g.,
        #     # This reads well.
        #     @beartyped pep_hinted() parameter pep_hinted_param=(1,) violates
        #     PEP type hint typing.Union[int, typing.Sequence[str]], as (1,):
        #     * Not int.
        #     * Tuple item 0 value "1" not str.
        #
        #     # This does not.
        #     @beartyped pep_hinted() parameter pep_hinted_param=(1,) violates
        #     PEP type hint typing.Union[int, typing.Sequence[str]], as (1,):
        #     * Tuple item 0 value "1" not str.
        #     * Not int.
        #
        # Note that prepending to lists is an O(n) operation, but that this
        # cost is negligible in this case both due to the negligible number of
        # child hints of the average "typing.Union" in general *AND* due to the
        # fact that this function is only called when a catastrophic type-check
        # failure has already occurred.
        cause_strs.insert(0, f'not {cause_types_unsatisfied}')
    # Else, this pith shallowly satisfies *ALL* the types of this union.

    # If prior logic appended *NO* causes, raise an exception.
    if not cause_strs:
        raise _BeartypeCallHintPepRaiseException(
            f'{cause.exception_prefix}type hint '
            f'{repr(cause.hint)} failure causes unknown.'
        )
    # Else, prior logic appended one or more strings describing these failures.

    # Output cause to be returned, permuted from this input cause such that the
    # output cause justification is either...
    cause_return = cause.permute_cause(cause_str_or_none=(
        # If prior logic appended one cause, a single-line
        # substring intended to be embedded in a longer string;
        f'{pith_repr} {cause_strs[0]}'
        if len(cause_strs) == 1 else
        # Else, prior logic appended two or more causes. In this case, a
        # multiline string comprised of...
        '{}:\n{}'.format(
            # This truncated object representation followed by...
            pith_repr,
            # The newline-delimited concatenation of each cause as a discrete
            # bullet-prefixed line...
            '\n'.join(
                '{}* {}'.format(
                    # Indented by the current indent...
                    cause.cause_indent,
                    # Whose first character is uppercased...
                    uppercase_str_char_first(
                        # Suffixed by a period if not yet suffixed by a period.
                        suffix_str_unless_suffixed(text=cause_str, suffix='.')
                    )
                )
                # '{}* {}.'.format(cause_indent, uppercase_str_char_first(cause_union))
                for cause_str in cause_strs
            )
        )
    ))

    # Return this cause.
    return cause_return



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Test-wide **import hookable LangChain integration submodule** (i.e., data module
defining LangChain-specific Runnables decorated by the third-party
decorator-hostile :func:`langchain_core.runnables.chain` decorator which the
:mod:`beartype.beartype` decorator will then be injected *after* rather than
*before* as the earliest decorator, mimicking real-world usage of
:func:`beartype.claw` import hooks from external callers).

See Also
--------
https://github.com/beartype/beartype/issues/541
    StackOverflow issue exercised by this submodule.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeCallHintParamViolation
from langchain_core.runnables import chain
from pytest import raises

# ....................{ FUNCTIONS                          }....................
@chain
def when_harboured_in(the_sleepy_west: dict[str, str]) -> str:
    '''
    Arbitrary function trivially satisfying LangChain's Runnable protocol by
    accepting an arbitrary dictionary of mock input data and returning a string
    mocking the output text response of a large language model (LLM) passed that
    input.

    The :obj:`.chain` decorator wraps this function in a LangChain-specific
    ``Runnable`` object defining an ``invoke()`` method, which we then call
    below to validate this wrapping.
    '''

    # Value associated with this key of this caller-defined mock input.
    upon_exalted_couch = the_sleepy_west['After the full completion']

    # Mock an LLM response to this mock input.
    return f'For rest divine {upon_exalted_couch}.'

# ....................{ PASS                               }....................
# Implicitly assert that the import hook registered by the caller respected the
# decorator-hostile @chain decorator decorating the function defined above by
# injecting the @beartype decorator as the last decorator on that function.
#
# If this call raises an unexpected exception, then that import hook failed to
# respect that the decorator-hostile @chain decorator by instead injecting the
# @beartype decorator as the first decorator on that function. Since @chain is
# decorator-hostile and thus hostile to @beartype as well, @chain prohibits
# *ANY* decorator from being applied after itself.
the_arms_of_melody = when_harboured_in.invoke(
    {'After the full completion': 'of fair day'})

# Explicitly assert that call returned the expected output.
assert the_arms_of_melody == 'For rest divine of fair day.'

# ....................{ FAIL                               }....................
# Assert this Runnable when passed an invalid parameter raises the expected
# @beartype-specific type-checking violation.
with raises(BeartypeCallHintParamViolation):
    when_harboured_in.invoke(
        {b"He pac'd away": b'the pleasant hours of ease'})



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`696`-compliant **type parameter default** (i.e., child type
hints specified as the default type hints to which type parameters with no other
concrete mappings default under Python >= 3.13) utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep696Exception
from beartype._data.kind.datakindiota import SENTINEL
from beartype._data.typing.datatyping import (
    Pep484612646TypeArgPacked,
    TypeException,
)
from beartype._data.typing.datatypingport import (
    # Hint,
    HintOrSentinel,
)
from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_13

# ....................{ GETTERS                            }....................
#FIXME: Unit test us up, please. *sigh*
def get_hint_pep484612646_typearg_packed_default_or_sentinel(
    # Mandatory parameters.
    hint: Pep484612646TypeArgPacked,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep696Exception,
    exception_prefix: str = '',
) -> HintOrSentinel:
    '''
    :pep:`696`-compliant **type parameter default** (i.e., child hint that may
    be an unquoted forward reference referring to a currently undefined type,
    initially passed as the value of the optional ``default`` parameter on the
    construction of the passed type parameter) defined on the passed **type
    parameter** (i.e., :pep:`484`-compliant type variable or
    :pep:`646`-compliant type variable tuple) if any *or* the sentinel
    placeholder otherwise (i.e., if no default was defined on this type
    parameter).

    This getter intentionally returns the sentinel placeholder (rather than
    :data:`None`) when either of these attributes is the null subhint value.
    Why? To enable callers to distinguish between type parameters lacking
    defaults and type parameters whose defaults are actually :data:`None`, which
    is (of course) a valid :pep:`484`-compliant type hint.

    This getter is intentionally *not* memoized (e.g., by the
    ``@callable_cached`` decorator), as the implementation trivially reduces to
    a one-liner.

    Parameters
    ----------
    hint : Pep484612646TypeArgPacked
        Type parameter to be inspected.
    exception_cls : TypeException, default: BeartypeDecorHintPep696Exception
        Type of exception to be raised in the event of a fatal error. Defaults
        to :exc:`.BeartypeDecorHintPep696Exception`.
    exception_prefix : str, default: ''
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    HintOrSentinel
        Either:

        * If this type parameter defines a default, that default.
        * Else, the sentinel placeholder.

    Raises
    ------
    exception_cls
        If the passed type hint is *not* a type parameter.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.proposal.pep484612646 import (
        die_unless_hint_pep484612646_typearg_packed)
    from beartype._util.hint.pep.proposal.pep749 import (
        get_hint_pep749_subhint_optional)

    # If this hint is *NOT* a packed type parameter, raise an exception.
    die_unless_hint_pep484612646_typearg_packed(
        hint=hint,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )
    # Else, this hint is a packed type parameter.

    # Child hint to which this type parameter defaults if this type parameter
    # has a default *OR* the sentinel placeholder otherwise.
    hint_default = SENTINEL

    # If the active Python interpreter targets Python >= 3.13 and thus supports
    # PEP 696.
    if IS_PYTHON_AT_LEAST_3_13:
        # Defer version-specific imports.
        from beartype.typing import NoDefault  # pyright: ignore

        # Child hint to which this type parameter defaults if this type
        # parameter has a default *OR* the sentinel placeholder otherwise.
        hint_default = get_hint_pep749_subhint_optional(
            hint=hint,  # pyright: ignore
            subhint_name_dynamic='evaluate_default',
            subhint_name_static='__default__',
            subhint_value_null=NoDefault,
            exception_cls=exception_cls,
            exception_prefix=exception_prefix,
        )
    # Else, the active Python interpreter targets Python < 3.12 and thus fails
    # to support PEP 696.

    # Return this type parameter default.
    return hint_default



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`- and :pep:`585`-compliant **generic type hint
testers** (i.e., low-level callables generically validating and detecting both
:pep:`484`- and :pep:`585`-compliant generic classes).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep484585Exception
from beartype._data.cls.datacls import TYPES_PEP484544_GENERIC
from beartype._data.typing.datatypingport import Hint
from beartype._data.typing.datatyping import TypeException
from beartype._util.bear.utilbearblack import is_object_blacklisted
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.hint.pep.proposal.pep484.pep484generic import (
    is_hint_pep484_generic_subbed,
    is_hint_pep484_generic_unsubbed,
)
from beartype._util.hint.pep.proposal.pep585 import (
    is_hint_pep585_generic_subbed,
    is_hint_pep585_generic_unsubbed,
    is_hint_pep585_builtin_subbed,
)

# ....................{ RAISERS                            }....................
def die_unless_hint_pep484585_generic_unsubbed(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    exception_cls: TypeException = BeartypeDecorHintPep484585Exception,
    exception_prefix: str = '',
) -> None:
    '''
    Raise the passed exception unless the passed type hint is an **unsubscripted
    generic** (i.e., type originally subclassing at least one subscripted
    :pep:`484`- or :pep:`585`-compliant pseudo-superclass).

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.
    exception_cls : TypeException
        Type of exception to be raised. Defaults to
        :exc:`.BeartypeDecorHintPep484585Exception`.
    exception_prefix : str, optional
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Raises
    ------
    exception_cls
        If this hint is *not* an unsubscripted generic.
    '''

    # If this hint is *NOT* an unsubscripted generic...
    if not is_hint_pep484585_generic_unsubbed(hint):
        assert isinstance(exception_cls, type), (
            f'{repr(exception_cls)} not type.')
        assert isinstance(exception_prefix, str), (
            f'{repr(exception_prefix)} not string.')

        # Raise an exception of this type prefixed by this prefix.
        raise exception_cls(
            f'{exception_prefix}type hint {repr(hint)} not '
            f'PEP 484 or 585 unsubscripted generic.'
        )
    # Else, this hint is an unsubscripted generic.

# ....................{ TESTERS                            }....................
def is_hint_pep484585_generic_user(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed :pep:`484`- or :pep:`585`-compliant generic
    is **user-defined** (i.e., defined by a third-party downstream codebase
    rather than CPython's first-party upstream standard library).

    Specifically, this tester returns :data:`True` only if this generic is
    neither:

    * A :pep:`484`- or :pep:`544`-compliant superclass defined by the
      :mod:`typing` module (e.g., :class:`typing.Generic`,
      :class:`typing.Protocol`) *nor*...
    * A :pep:`585`-compliant superclass (e.g., ``list[T]``).

    This tester is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as the implementation trivially reduces to
    an efficient one-liner.

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this is a user-defined generic.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepget import (
        get_hint_pep_origin_type_or_none)

    # Return true only if...
    return (
        # This object is a generic that is neither...
        is_hint_pep484585_generic(hint) and not (
            # A PEP 585-compliant subscripted superclass (e.g., "list[T]")
            # *NOR*...
            is_hint_pep585_builtin_subbed(hint) or

            #FIXME: *YIKES.* This approach omits "typing_extensions.Protocol",
            #which is a distinct type from "typing.Protocol". Instead of
            #dogmatically testing raw types, consider a more flexible approach
            #that portably tests higher-level names: e.g.,
            #    hint_origin = get_hint_pep_origin_type_or_none(
            #        hint=hint,
            #        # Preserve "typing.Generic" and "typing.Protocol" as themselves,
            #        # as doing so dramatically simplifies this test. *shrug*
            #        is_self_fallback=True,
            #    )
            #    if (
            #        is_hint_pep_type_typing(hint_origin) and
            #        hint_origin.__name__ in frozenset(('Generic', 'Protocol',))
            #    ):
            #        return False
            #FIXME: *HMM.* The above approach is certainly better, but still
            #falls short. Below, we suggest that the PEP 484-compliant
            #subscripted pseudo-superclass "typing.Generic[S]" should also be
            #seen as *NOT* a user-defined generic. Makes sense... except that
            #the above logic erroneously returns "True" for "typing.Generic[S]"!
            #Clearly, we also need to discard subscriptions. *sigh*
            #FIXME: Excise "TYPES_PEP484544_GENERIC" entirely. That tuple is
            #implicitly non-portable and thus bad news all around, really.

            # A subscripted or unsubscripted PEP 484- or 544-compliant
            # superclass defined by the standard "typing" module, including:
            # * "typing.Generic".
            # * "typing.Generic[S]".
            # * "typing.Protocol".
            # * "typing.Protocol[S]".
            get_hint_pep_origin_type_or_none(
                hint=hint,
                # Preserve "typing.Generic" and "typing.Protocol" as themselves,
                # as doing so dramatically simplifies this test. *shrug*
                is_self_fallback=True,
            ) in TYPES_PEP484544_GENERIC
        )
    )

# ....................{ TESTERS ~ kind                     }....................
@callable_cached
def is_hint_pep484585_generic(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed object is either a :pep:`484`- or
    :pep:`585`-compliant **generic** (i.e., either a type originally subclassing
    at least one subscripted :pep:`484`- or :pep:`585`-compliant
    pseudo-superclass *or* an object subscripted by one or more child type hints
    originating from such a type).

    This tester returns :data:`True` only if this object is either:

    * A :pep:`484`-compliant generic as tested by the lower-level
      :func:`.is_hint_pep484_generic` function.
    * A :pep:`585`-compliant generic as tested by the lower-level
      :func:`.is_hint_pep585_generic` function.

    This tester is memoized for efficiency.

    Caveats
    -------
    **Generics are not necessarily classes,** despite originally being declared
    as classes. Although *most* generics are classes, subscripting a generic
    class usually produces a generic non-class that *must* nonetheless be
    transparently treated as a generic class: e.g.,

    .. code-block:: pycon

       >>> from typing import Generic, TypeVar
       >>> S = TypeVar('S')
       >>> T = TypeVar('T')
       >>> class MuhGeneric(Generic[S, T]): pass
       >>> non_class_generic = MuhGeneric[S, T]
       >>> isinstance(non_class_generic, type)
       False

    Parameters
    ----------
    hint : Hint
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a generic.

    See Also
    --------
    :func:`beartype._util.hint.pep.utilpepget.get_hint_pep_typeargs_packed`
        Commentary on the relation between generics and parametrized hints.
    '''

    # Return true only if...
    return (
        # This hint is either a...
        (
            # PEP 484-compliant generic *OR*...
            #
            # Note these tests trivially reduce to fast O(1) operations and are
            # thus tested first.
            is_hint_pep484_generic_unsubbed(hint) or
            is_hint_pep484_generic_subbed(hint) or
            # PEP 585-compliant generic.
            #
            # Note this test is O(n) for n the number of pseudo-superclasses
            # originally subclassed by this generic and is thus tested last.
            is_hint_pep585_generic_unsubbed(hint) or
            is_hint_pep585_generic_subbed(hint)
        # *AND*...
        ) and
        # This generic is *NOT* beartype-blacklisted.
        not _is_hint_pep484585_generic_blacklisted(hint)
    )


def is_hint_pep484585_generic_subbed(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed object is either a :pep:`484`- or
    :pep:`585`-compliant **subscripted generic** (i.e., object subscripted by
    one or more child type hints originating from a type originally subclassing
    at least one subscripted :pep:`484`- or :pep:`585`-compliant
    pseudo-superclass).

    This tester returns :data:`True` only if this object is either:

    * A :pep:`484`-compliant subscripted generic as tested by the lower-level
      :func:`.is_hint_pep484_generic_subbed` function.
    * A :pep:`585`-compliant subscripted generic as tested by the lower-level
      :func:`.is_hint_pep585_generic_subbed` function.

    This tester is memoized for efficiency.

    Parameters
    ----------
    hint : Hint
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a subscripted generic.
    '''

    # Return true only if...
    return (
        # This hint is either a...
        (
            # PEP 484-compliant subscripted generic *OR*...
            #
            # Note this test trivially reduces to a fast O(1) operation and is
            # thus tested first.
            is_hint_pep484_generic_subbed(hint) or
            # PEP 585-compliant subscripted generic.
            #
            # Note this test is O(n) for n the number of pseudo-superclasses
            # originally subclassed by this generic and is thus tested last.
            is_hint_pep585_generic_subbed(hint)
        # *AND*...
        ) and
        # This generic is *NOT* beartype-blacklisted.
        not _is_hint_pep484585_generic_blacklisted(hint)
    )


def is_hint_pep484585_generic_unsubbed(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed object is either a :pep:`484`- or
    :pep:`585`-compliant **unsubscripted generic** (i.e., type originally
    subclassing at least one subscripted :pep:`484`- or :pep:`585`-compliant
    pseudo-superclass).

    This tester returns :data:`True` only if this object is either:

    * A :pep:`484`-compliant unsubscripted generic as tested by the lower-level
      :func:`.is_hint_pep484_generic_unsubbed` function.
    * A :pep:`585`-compliant unsubscripted generic as tested by the lower-level
      :func:`.is_hint_pep585_generic_unsubbed` function.

    This tester is memoized for efficiency.

    Parameters
    ----------
    hint : Hint
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a unsubscripted generic.
    '''

    # Return true only if...
    return (
        # This hint is either a...
        (
            # PEP 484-compliant unsubscripted generic *OR*...
            #
            # Note this test trivially reduces to a fast O(1) operation and is
            # thus tested first.
            is_hint_pep484_generic_unsubbed(hint) or
            # PEP 585-compliant unsubscripted generic.
            #
            # Note this test is O(n) for n the number of pseudo-superclasses
            # originally subclassed by this generic and is thus tested last.
            is_hint_pep585_generic_unsubbed(hint)
        # *AND*...
        ) and
        # This generic is *NOT* beartype-blacklisted.
        not _is_hint_pep484585_generic_blacklisted(hint)
    )

# ....................{ PRIVATE ~ testers                  }....................
#FIXME: Call above, please.
@callable_cached
def _is_hint_pep484585_generic_blacklisted(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed :pep:`484`- or :pep:`585`-compliant generic
    is **beartype-blacklisted** (i.e., defined in a third-party package or
    module known to be hostile to runtime type-checking).

    This tester is memoized for efficiency.

    Parameters
    ----------
    hint : Hint
        Generic to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this generic is beartype-blacklisted.

    See Also
    --------
    :func:`.is_object_blacklisted`
        Further details on beartype-blacklisting.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget import (
        get_hint_pep484585_generic_type)

    # Either:
    # * If this generic is already unsubscripted, this generic as is.
    # * Else, this generic is subscripted. In this case, the unsubscripted
    #   generic underlying this subscripted generic.
    hint_type = get_hint_pep484585_generic_type(hint)

    # For each possibly erased superclass of this generic, arbitrarily iterated
    # according to the method resolution order (MRO) for this generic...
    for hint_base in hint_type.__mro__:
        # If this superclass is beartype-blacklisted (i.e., defined in a
        # third-party package or module known to be hostile to runtime
        # type-checking), return true immediately.
        if is_object_blacklisted(hint_base):
            return True
        # Else, this superclass is *NOT* beartype-blacklisted. In this case,
        # continue to the next such superclass of this generic.
    # Else, all superclasses of this generic are *NOT* beartype-blacklisted.

    # Return false as a sane fallback.
    return False



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`-compliant **union type hint utilities** (i.e.,
callables generically applicable to :pep:`484`-compliant :attr:`typing.Union`
subscriptions).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep484Exception
from beartype.typing import Union
from beartype._data.typing.datatypingport import (
    Hint,
    TupleHints,
)
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_14

# ....................{ FACTORIES                          }....................
@callable_cached
def make_hint_pep484_union(hints: TupleHints) -> Hint:
    '''
    :pep:`484`-compliant **union type hint** (:attr:`typing.Union`
    subscription) synthesized from the passed tuple of two or more
    PEP-compliant type hints if this tuple contains two or more items, the one
    PEP-compliant type hint in this tuple if this tuple contains only one item,
    *or* raise an exception otherwise (i.e., if this tuple is empty).

    This factory is memoized for efficiency. Technically, the
    :attr:`typing.Union` type hint factory already caches its subscripted
    arguments. Pragmatically, that caching is slow and thus worth optimizing
    with trivial optimization on our end. Moreover, this factory is called by
    the performance-sensitive
    :func:`beartype._check.convert._convcoerce.coerce_hint_any` coercer in an
    early-time code path of the :func:`beartype.beartype` decorator. Optimizing
    this factory thus optimizes :func:`beartype.beartype` itself.

    Parameters
    ----------
    hint : TupleHints
        Type hint to be inspected.

    Returns
    -------
    Hint
        Either:

        * If this tuple contains two or more items, the union type hint
          synthesized from these items.
        * If this tuple contains only one item, this item as is.

    Raises
    ------
    BeartypeDecorHintPep484Exception
        If this tuple is empty.
    '''
    assert isinstance(hints, tuple), f'{repr(hints)} not tuple.'

    # If this tuple is empty, raise an exception.
    if not hints:
        raise BeartypeDecorHintPep484Exception('"hints" tuple empty.')
    # Else, this tuple contains one or more child type hints.

    # Return either...
    return (
        # If the active Python interpreter targets Python >= 3.14, the PEP
        # 484-compliant union dynamically created by deferring to the C-based
        # typing.Union.__class_getitem__() class method.
        #
        # Note that this method does *NOT* exist under older Python versions.
        Union.__class_getitem__(hints)  # type: ignore[attr-defined]
        if IS_PYTHON_AT_LEAST_3_14 else
        # Else, the active Python interpreter targets Python <= 3.13. In this
        # case, the PEP 484-compliant union dynamically created by deferring to
        # the pure-Python typing.Union.__getitem__() instance method.
        #
        # Note that this method still exists but is *NOT* safely callable under
        # newer Python versions, where doing so raises "TypeError" exceptions
        # resembling:
        #     TypeError: descriptor '__getitem__' requires a 'typing.Union'
        #     object but received a 'tuple'
        Union.__getitem__(hints)  # type: ignore[return-value]
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **warning handlers** (i.e., low-level callables manipulating
non-fatal warnings -- which, technically, are also exceptions -- in a
human-readable, general-purpose manner).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    Any,
    Iterable,
    Iterator,
)
from beartype._data.error.dataerrmagic import EXCEPTION_PLACEHOLDER
from beartype._data.typing.datatyping import TypeWarning
from beartype._util.error.utilerrtest import is_exception_message_str
from beartype._util.py.utilpyversion import (
    IS_PYTHON_AT_LEAST_3_11,
    IS_PYTHON_AT_LEAST_3_12,
)
from beartype._util.text.utiltextmunge import uppercase_str_char_first
from collections.abc import Iterable as IterableABC
from contextlib import contextmanager
from warnings import (
    WarningMessage,
    catch_warnings,
    simplefilter,
    warn,
    warn_explicit,
)

# ....................{ CONTEXTS                           }....................
#FIXME: Unit test us up, please.
@contextmanager
def warnings_ignored() -> Iterator[None]:
    '''
    Context manager temporarily ignoring *all* warnings transitively emitted
    within the body of this context.

    Yields
    ------
    None
        This context manager yields *no* objects.

    See Also
    --------
    https://stackoverflow.com/a/14463362/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''

    # If the active Python interpreter targets Python > 3.11, prefer an
    # efficient one-liner yielding the desired outcome. Get it? Yielding? ...heh
    if IS_PYTHON_AT_LEAST_3_11:
        with catch_warnings(action='ignore'):  # type: ignore[call-overload]
            yield
    # Else, the active Python interpreter targets Python <= 3.10. In this case,
    # fallback to an inefficient generator yielding the same outcome.
    else:
        with catch_warnings():
            simplefilter('ignore')
            yield

# ....................{ WARNERS                            }....................
# If the active Python interpreter targets Python >= 3.12, the standard
# warnings.warn() function supports the optional "skip_file_prefixes" parameter
# critical for emitting more useful warnings. In this case, define the
# issue_warning() warner to pass that parameter.
if IS_PYTHON_AT_LEAST_3_12:
    # ....................{ IMPORTS                        }....................
    # Defer version-specific imports.
    import beartype
    from os.path import dirname

    # ....................{ WARNERS                        }....................
    def issue_warning(cls: TypeWarning, message: str) -> None:
        # The warning you gave us is surely our last!
        warn(message, cls, skip_file_prefixes=_ISSUE_WARNING_IGNORE_DIRNAMES)  # type: ignore[call-overload]
        # warn(message, cls)  # type: ignore[call-overload]

    # ....................{ PRIVATE ~ globals              }....................
    _ISSUE_WARNING_IGNORE_DIRNAMES = (dirname(beartype.__file__),)
    '''
    Tuple of one or more **ignorable warning dirnames** (i.e., absolute
    directory names of all Python modules to be ignored by the
    :func:`.issue_warning` warner when deciding which source module to associate
    with the issued warning, enabling this warner to associate this warning with
    the original externally defined module to which this warning applies).

    This tuple includes the dirname of the top-level directory providing the
    :mod:`beartype` package, enabling this warner to ignore all stack frames
    produced by internal calls to submodules of this package. Doing so emits
    substantially more useful and readable warnings for external callers.
    '''
# Else, the active Python interpreter targets Python < 3.12. In this case,
# define the issue_warning() warner to avoid passing that parameter.
else:
    def issue_warning(cls: TypeWarning, message: str) -> None:
        # Time to cry your tears! Now cry!
        warn(message, cls)


issue_warning.__doc__ = (
    '''
    Issue (i.e., emit) a non-fatal warning of the passed type with the passed
    message.

    Caveats
    -------
    **This high-level warner should always be called in lieu of the low-level**
    :func:`warnings.warn` **warner.** Whereas the latter issues warnings that
    obfuscate the external user-defined modules to which those warnings apply,
    this warner associates this warning with the applicable user-defined module
    when the active Python interpreter targets Python >= 3.12.

    Parameters
    ----------
    cls: Type[Warning]
        Type of warning to be issued.
    message: str
        Human-readable warning message to be issued.
    '''
)

# ....................{ REWARNERS                          }....................
def reissue_warnings_placeholder(
    # Mandatory parameters.
    warnings: Iterable[WarningMessage],
    target_str: str,

    # Optional parameters.
    source_str: str = EXCEPTION_PLACEHOLDER,
) -> None:
    '''
    Reissue (i.e., re-emit) the passed warning in a safe manner preserving both
    this warning object *and* **associated context** (e.g., filename, line
    number) associated with this warning object, but globally replacing all
    instances of the passed source substring hard-coded into this warning's
    message with the passed target substring.

    Parameters
    ----------
    warnings : Iterable[WarningMessage]
        Iterable of zero or more warnings to be reissued, typically produced by
        an external call to the standard
        ``warnings.catch_warnings(record=True)`` context manager.
    target_str : str
        Target human-readable format substring to replace the passed source
        substring previously hard-coded into this warning's message.
    source_str : Optional[str]
        Source non-human-readable substring previously hard-coded into this
        warning's message to be replaced by the passed target substring.
        Defaults to :data:`.EXCEPTION_PLACEHOLDER`.

    Warns
    -----
    warning
        The passed warning, globally replacing all instances of this source
        substring in this warning's message with this target substring.

    See Also
    --------
    :data:`.EXCEPTION_PLACEHOLDER`
        Further commentary on usage and motivation.
    https://stackoverflow.com/a/77516994/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''
    assert isinstance(warnings, IterableABC), (
        f'{repr(warnings)} not iterable.')
    assert isinstance(source_str, str), f'{repr(source_str)} not string.'
    assert isinstance(target_str, str), f'{repr(target_str)} not string.'

    # For each warning descriptor in this iterable of zero or more warning
    # descriptors...
    for warning_info in warnings:
        assert isinstance(warning_info, WarningMessage), (  # <-- terrible name!
           f'{repr(warning_info)} not "WarningMessage" instance.')

        # Original warning wrapped by this warning descriptor, localized both
        # for readability *AND* negligible speed. *sigh*
        warning = warning_info.message

        # Munged warning message to be issued below.
        warning_message_new: Any = None

        # If either...
        if (
            # This warning is... *ALREADY A STRING!?* What is going on here?
            # Look. Not even we know. But mypy claims that warnings recorded by
            # calls to the standard "warnings.catch_warnings(record=True)"
            # function satisfy the union "Warning | str". Technically, that
            # makes no sense. Pragmatically, that makes no sense. But mypy says
            # it's true. We are too tired to argue with static type-checkers at
            # 4:11AM in the morning.
            isinstance(warning, str) or
            # This warning is conventional...
            is_exception_message_str(warning)
        # Then this warning is or has a standard message. In this case...
        ):
            # Original warning message, coerced from the original warning.
            #
            # Note that the poorly named "message" attribute is the original
            # warning rather warning message. Just as with exceptions, coercing
            # this warning into a string reliably retrieves its message.
            warning_message_old = str(warning)

            # Munged warning message globally replacing all instances of this
            # source substring with this target substring.
            #
            # Note that we intentionally call the lower-level str.replace()
            # method rather than the higher-level
            # beartype._util.text.utiltextmunge.replace_str_substrs() function
            # here, as the latter unnecessarily requires this warning message to
            # contain one or more instances of this source substring.
            warning_message_new = warning_message_old.replace(
                source_str, target_str)

            # If doing so actually changed this message...
            if warning_message_new != warning_message_old:
                # Uppercase the first character of this message if needed.
                warning_message_new = uppercase_str_char_first(
                    warning_message_new)
            # Else, this message remains preserved as is.
        # Else, this is an unconventional warning. In this case...
        else:
            # Tuple of the zero or more arguments with which this warning was
            # originally issued.
            warning_args = warning.args

            # Assert that this warning was issued with exactly one argument.
            # Since the warnings.warn() signature accepts only a single
            # "message" parameter, this assertion *SHOULD* always hold. *sigh*
            assert len(warning_args) == 1

            # Preserve this warning as is.
            warning_message_new = warning_args[0]

        # Reissue this warning with a possibly modified message.
        warn_explicit(
            message=warning_message_new,
            category=warning_info.category,
            filename=warning_info.filename,
            lineno=warning_info.lineno,
            source=warning_info.source,
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator** :pep:`613`-compliant **unit tests**.

This submodule unit tests :pep:`613` support implemented in the
:func:`beartype.beartype` decorator.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_if_python_version_less_than

# ....................{ TESTS                              }....................
@skip_if_python_version_less_than('3.10.0')
def test_decor_pep613() -> None:
    '''
    Test :pep:`613` support implemented in the :func:`beartype.beartype`
    decorator if the active Python interpreter targets Python >= 3.10 *or* skip
    otherwise.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeDecorHintPep613DeprecationWarning
    from beartype.typing import TypeAlias
    from pytest import warns
    from warnings import simplefilter

    # ....................{ MAIN                           }....................
    # Force pytest to temporarily allow deprecation warnings to be caught by the
    # warns() context manager for the duration of this test. By default, pytest
    # simply "passes through" all deprecation warnings for subsequent reporting
    # if tests otherwise successfully pass. Deprecation warnings include:
    # * "DeprecationWarning".
    # * "FutureWarning".
    # * "PendingDeprecationWarning".
    simplefilter('always')

    # ....................{ CALLABLES                      }....................
    with warns(BeartypeDecorHintPep613DeprecationWarning):
        @beartype
        def and_shook_him() -> TypeAlias:
            '''
            Arbitrary callable annotated by the deprecated :pep:`613`-compliant
            :obj:`typing.TypeAlias` type hint singleton, which is shallowly
            ignorable and thus satisfied by *all* possible objects.
            '''

            return 'And shook him from his rest, and led him forth'

    # ....................{ ASSERTS                        }....................
    # Assert that calling a PEP 613-annotated callable succeeds as expected.
    assert and_shook_him() == 'And shook him from his rest, and led him forth'



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`646` and :pep:`692` **type hint utility** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep646692` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ tester                     }....................
def test_is_hint_pep646_tuple_unpacked_prefix() -> None:
    '''
    Test the private
    :mod:`beartype._util.hint.pep.proposal.pep646692.is_hint_pep646_tuple_unpacked_prefix`
    tester.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.hint.pep.proposal.pep646692 import (
        is_hint_pep646_tuple_unpacked_prefix,
        make_hint_pep646_tuple_unpacked_prefix,
    )
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_11

    # ....................{ VERSION                        }....................
    # If the active Python interpreter targets Python >= 3.11 and thus supports
    # PEP 646...
    if IS_PYTHON_AT_LEAST_3_11:
        # PEP 646-compliant unpacked child tuple hint subscripted by arbitrary
        # child child hints.
        hint_pep646_tuple_unpacked = make_hint_pep646_tuple_unpacked_prefix(
            (int, float))

        # Assert this tester accepts this hint.
        assert is_hint_pep646_tuple_unpacked_prefix(
            hint_pep646_tuple_unpacked) is True
    # Else, the active Python interpreter targets Python <= 3.10 and thus fails
    # to support PEP 646.

    # ....................{ FAIL                           }....................
    # Assert this tester rejects PEP 484- and 585-compliant tuple type hints of
    # both fixed- and variable-length variants.
    assert is_hint_pep646_tuple_unpacked_prefix(tuple[int, ...]) is False
    assert is_hint_pep646_tuple_unpacked_prefix(tuple[bool, int, float]) is False

    # Assert this tester rejects unrelated arbitrary objects.
    assert is_hint_pep646_tuple_unpacked_prefix(
        'The Titans fierce, self-hid, or prison-bound,') is False



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **object attribute utility** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.utilobjattr` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ getter                     }....................
def test_get_object_attrs_name_to_value_explicit() -> None:
    '''
    Test both the
    :func:`beartype._util.utilobjattr.get_object_attrs_name_to_value_explicit`
    *and*
    :func:`beartype._util.utilobjattr.get_object_methods_name_to_value_explicit`
    getters, whose testing requires identical types and objects.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.utilobjattr import (
        get_object_attrs_name_to_value_explicit,
        get_object_methods_name_to_value_explicit,
    )
    from beartype._util.text.utiltextidentifier import is_dunder

    # ....................{ CLASSES                        }....................
    class StrongShuddering(object):
        '''
        Arbitrary superclass.
        '''

        def from_his_burning_limbs(self) -> None:
            '''
            Arbitrary method.
            '''

        as_one: str = 'Strong shuddering from his burning limbs. As one'
        '''
        Arbitrary class variable.
        '''


    class RousedBy(StrongShuddering):
        '''
        Arbitrary subclass.
        '''

        def some_joyous_madness(self) -> None:
            '''
            Arbitrary method.
            '''

        @property
        def from_the_couch(self) -> None:
            '''
            Arbitrary property getter.
            '''

            return None

    # ....................{ LOCALS                         }....................
    # Arbitrary instance of this subclass.
    of_fever = RousedBy()

    # ....................{ PASS ~ attribute               }....................
    # Assert this getter returns the expected dictionary when passed this
    # instance *WITHOUT* a predicate. Note that asserting the exact contents of
    # this dictionary is non-trivial across Python versions and thus avoided,
    # due to this dictionary containing *ALL* slot wrappers implicitly inherited
    # from the root "object" superclass (e.g., object.__str__()).
    of_fever_attrs_expect = {
        'as_one': of_fever.as_one,
        'from_his_burning_limbs': StrongShuddering.from_his_burning_limbs,
        'some_joyous_madness': RousedBy.some_joyous_madness,
        'from_the_couch': type(of_fever).__dict__['from_the_couch'],
    }
    of_fever_attrs_actual = get_object_attrs_name_to_value_explicit(of_fever)
    assert of_fever_attrs_expect.items() <= of_fever_attrs_actual.items()

    # Assert this getter returns the expected dictionary when passed this
    # instance *WITH* a predicate.
    assert get_object_attrs_name_to_value_explicit(
        obj=of_fever,
        predicate=lambda attr_name, attr_value: not is_dunder(attr_name)
    ) == of_fever_attrs_expect

    # ....................{ PASS ~ method                  }....................
    # Assert this getter returns the expected dictionary when passed this
    # instance *WITHOUT* a predicate. Note that asserting the exact contents of
    # this dictionary is non-trivial across Python versions and thus avoided.
    # Although asserting the contents of this dictionary was trivial under
    # Python <= 3.13, Python >= 3.14 complicates comparisons by introducing
    # dynamically instantiated bound methods for certain contexts, including:
    # * The __annotate_func__() dunder method, which CPython now implicitly
    #   defines for *ALL* types declaring one or more class variables annotated
    #   by type hints.
    of_fever_methods_expect = {
        'from_his_burning_limbs': StrongShuddering.from_his_burning_limbs,
        'some_joyous_madness': RousedBy.some_joyous_madness,
    }
    of_fever_methods_actual = (
        get_object_methods_name_to_value_explicit(of_fever))
    assert of_fever_methods_expect.items() <= of_fever_methods_actual.items()

    # ....................{ PASS ~ method : dir            }....................
    # List of the names of all attributes bound to this object.
    of_fever_dir = dir(of_fever)

    # Remove an arbitrary method name from this list.
    of_fever_dir.remove('some_joyous_madness')
    of_fever_methods_expect.pop('some_joyous_madness')

    # Assert this getter returns the expected dictionary when passed this
    # instance with an arbitrary method name removed from this list.
    of_fever_methods_actual = get_object_methods_name_to_value_explicit(
        obj=of_fever, obj_dir=of_fever_dir)
    assert of_fever_methods_expect.items() <= of_fever_methods_actual.items()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **type hint sign mappings** (i.e., dictionary globals mapping from
instances of the :class:`beartype._data.hint.sign.datahintsigncls.HintSign`
class to various metadata associated with categories of type hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import Dict
from beartype._data.api.standard.datatyping import TYPING_MODULE_NAMES
from beartype._data.typing.datatyping import (
    DictStrToHintSign,
)
from beartype._data.hint.sign.datahintsigncls import HintSign
from beartype._data.hint.sign.datahintsigns import (
    HintSignAbstractSet,
    HintSignAsyncContextManager,
    HintSignAsyncGenerator,
    HintSignAsyncIterator,
    HintSignAsyncIterable,
    HintSignAwaitable,
    HintSignChainMap,
    HintSignCollection,
    HintSignContainer,
    HintSignContextManager,
    HintSignCoroutine,
    HintSignCounter,
    HintSignDefaultDict,
    HintSignDeque,
    HintSignDict,
    HintSignForwardRef,
    HintSignFrozenSet,
    HintSignGenerator,
    HintSignItemsView,
    HintSignIterable,
    HintSignIterator,
    HintSignKeysView,
    HintSignList,
    HintSignMapping,
    HintSignMappingView,
    HintSignMatch,
    HintSignMutableMapping,
    HintSignMutableSequence,
    HintSignMutableSet,
    HintSignNamedTuple,
    HintSignNewType,
    HintSignNone,
    HintSignOrderedDict,
    HintSignParamSpec,
    HintSignParamSpecArgs,
    HintSignParamSpecKwargs,
    HintSignPattern,
    HintSignPep484585GenericUnsubbed,
    HintSignPep557DataclassInitVar,
    HintSignPep695TypeAliasUnsubscripted,
    HintSignProtocol,
    HintSignReversible,
    HintSignSequence,
    HintSignSet,
    HintSignTuple,
    HintSignType,
    HintSignTypedDict,
    HintSignTypeVar,
    HintSignTypeVarTuple,
    HintSignUnion,
    HintSignUnpack,
    HintSignValuesView,
)

# ....................{ HINTS                              }....................
HintSignTrie = Dict[str, DictStrToHintSign]
'''
PEP-compliant type hint matching a **hint sign trie** (i.e.,
dictionary-of-dictionaries tree data structure mapping from the fully-qualified
names of packages and modules to nested dictionaries mapping from the
unqualified basenames of various PEP-compliant type hints residing in those
packages and modules to their identifying signs).
'''

# ....................{ PRIVATE ~ constructors             }....................
def _init_hint_sign_trie(hint_sign_trie: HintSignTrie) -> HintSignTrie:
    '''
    Initialize the passed **hint sign trie** (i.e., dictionary-of-dictionaries
    tree data structure mapping from the fully-qualified names of packages and
    modules to nested dictionaries mapping from the unqualified basenames of
    various PEP-compliant type hints residing in those packages and modules to
    their identifying signs).

    Parameters
    ----------
    hint_sign_trie : HintSignTrie
        Hint sign trie to be initialized.

    Returns
    -------
    HintSignTrie
        This same trie as a caller convenience.
    '''

    # For the fully-qualified name of each quasi-standard typing module...
    for typing_module_name in TYPING_MODULE_NAMES:
        # If this trie fails to map this name to a nested dictionary, map this
        # name to the empty nested dictionary. Doing so simplifies logic
        # performed by the _init() function called below.
        if typing_module_name not in hint_sign_trie:
            hint_sign_trie[typing_module_name] = {}

    # Return this same trie as a caller convenience.
    return hint_sign_trie

# ....................{ MAPPINGS                           }....................
# The majority of this dictionary is initialized with automated inspection
# iterating over the "_TYPING_ATTR_HINT_BASENAME_TO_SIGN" dictionary local
# defined by the _init() function below. The *ONLY* key-value pairs explicitly
# defined here are those *NOT* amenable to such inspection.
#
# Note that the root "object" superclass *CANNOT* be safely mapped to the
# "typing.Any" singleton with logic resembling:
#    # ..................{ BUILTINS                           }..................
#    # Standard builtins module.
#    'builtins': {
#        # ..................{ NON-PEP                        }..................
#        # The PEP-noncompliant root "object" superclass is semantically
#        # equivalent to the PEP-compliant "typing.Any" singleton from the
#        # runtime type-checking perspective. Why? Because "object" is the
#        # transitive superclass of all classes. Attributes annotated as "object"
#        # unconditionally match *ALL* objects under isinstance()-based type
#        # covariance and thus semantically reduce to unannotated attributes.
#        # Reduce this hint to "typing.Any", which then reduces this hint to the
#        # ignorable "HINT_SANE_IGNORABLE" singleton.
#        'object': HintSignAny,
#    },
#
# Why? Because doing so would then erroneously render the otherwise
# PEP-noncompliant "object" type PEP-compliant, which would then cause
# raisers like die_if_pep() and die_unless_pep() to exhibit anomalous
# behaviour with respect to that type. In short, the clever solution is the
# wrong solution. *sigh*
HINT_MODULE_NAME_TO_HINT_BASENAME_TO_SIGN = _init_hint_sign_trie({})
'''
**Type hint trie** (i.e., dictionary-of-dictionaries tree data structure mapping
from the fully-qualified names of packages and modules to nested dictionaries
mapping from the unqualified basenames of all PEP-compliant type hints residing
in those packages and modules such that those hints are either callables or
classes *and* uniquely identifiable by those callables or classes to their
identifying signs).
'''


# The majority of this dictionary is initialized with automated inspection
# iterating over the "_TYPING_ATTR_TYPE_BASENAME_TO_SIGN" dictionary local
# defined by the _init() function below. The *ONLY* key-value pairs explicitly
# defined here are those *NOT* amenable to such inspection.
HINT_MODULE_NAME_TO_TYPE_BASENAME_TO_SIGN = _init_hint_sign_trie({
    # ..................{ BUILTINS                           }..................
    # Standard builtins module.
    'builtins': {
        # ..................{ PEP 484                        }..................
        # PEP 484-compliant forward reference type hints may be annotated
        # either:
        # * Explicitly as "typing.ForwardRef" instances, which automated
        #   inspection performed by the _init() function below already handles.
        # * Implicitly as strings, which this key-value pair here detects. Note
        #   this unconditionally matches *ALL* strings, including both:
        #   * Invalid Python identifiers (e.g., "0d@yw@r3z").
        #   * Absolute forward references (i.e., fully-qualified classnames)
        #     technically non-compliant with PEP 484 but seemingly compliant
        #     with PEP 585.
        #
        #   Since the distinction between PEP-compliant and -noncompliant
        #   forward references is murky at best and since unconditionally
        #   matching *ALL* string as PEP-compliant substantially simplifies
        #   logic throughout the codebase, we (currently) opt to do so.
        'str': HintSignForwardRef,

        # The C-based "builtins.NoneType" type does *NOT* actually exist: e.g.,
        #     >>> from builtins import NoneType
        #     ImportError: cannot import name 'NoneType' from 'builtins'
        #     (unknown location)
        #
        # This implies that users *CANNOT* make user-defined instances of this
        # type, which then implies that the *ONLY* instance of this type is
        # guaranteed to be the PEP 484-compliant "None" singleton, which
        # circuitously reduces to "types.NoneType" under PEP 484.
        #
        # PEP 484 explicitly supports this singleton as follows:
        #     When used in a type hint, the expression None is considered
        #     equivalent to type(None).
        #
        # Note that the representation of the type of the "None" singleton
        # (i.e., "<class 'NoneType'>") is intentionally omitted here despite the
        # "None" singleton reducing to that type. Indeed, the *ONLY* reason we
        # detect this singleton at all is to enable that reduction. Although
        # this singleton conveys a PEP-compliant semantic, the type of this
        # singleton explicitly conveys *NO* PEP-compliant semantics. That type
        # is simply a standard isinstanceable type (like any other). Indeed,
        # attempting to erroneously associate the type of the "None" singleton
        # with the same sign here would cause that type to be detected as
        # conveying sign-specific PEP-compliant semantics rather than *NO* such
        # semantics, which would then substantially break and complicate dynamic
        # code generation for no benefit whatsoever.
        'NoneType': HintSignNone,
    },

    # ..................{ ANNOTATIONLIB                      }..................
    # Standard PEP 649- and 749-compliant annotation module.
    'annotationlib': {
        # ..................{ PEP (649|749)                  }..................
        # The "annotationlib.ForwardRef" type is now the canonical location of
        # this type under Python >= 3.14, which previously resided at
        # "typing.ForwardRef". The latter still exists but simply as a lazy
        # shallow alias of the former.
        'ForwardRef': HintSignForwardRef,
    },

    # ..................{ DATACLASSES                        }..................
    # Standard PEP 557-compliant dataclass module.
    'dataclasses': {
        # ..................{ PEP 557                        }..................
        # PEP 557-compliant "dataclasses.InitVar" type hints are merely
        # instances of that class.
        'InitVar': HintSignPep557DataclassInitVar,
    },

    # ..................{ TYPES                              }..................
    # Standard module containing common low-level C-based types.
    'types': {
        #FIXME: Excise this *AFTER* dropping Python 3.13 support, please.
        # ..................{ PEP 604                        }..................
        # Python <= 3.13 implements PEP 604-compliant |-style unions (e.g., "int
        # | float") as instances of the low-level C-based "types.UnionType"
        # type. Thankfully, these unions are semantically interchangeable with
        # comparable PEP 484-compliant unions (e.g., "typing.Union[int,
        # float]"); both kinds expose equivalent dunder attributes (e.g.,
        # "__args__", "__parameters__"), enabling subsequent code generation to
        # conflate the two without issue.
        'UnionType': HintSignUnion,
    },

    # ..................{ TYPING                             }..................
    # Standard typing module.
    'typing': {
        # ..................{ PEP 484                        }..................
        # Python >= 3.10 implements PEP 484-compliant "typing.NewType" type
        # hints as instances of that pure-Python class.
        #
        # Note that we intentionally omit both "beartype.typing.NewType" *AND*
        # "typing_extensions.NewType" here, as:
        # * "beartype.typing.NewType" is a merely an alias of "typing.NewType".
        # * Regardless of the current Python version,
        #   "typing_extensions.NewType" type hints remain implemented in the
        #   manner of Python < 3.10 -- which is to say, as closures of that
        #   function. See also:
        #       https://github.com/python/typing/blob/master/typing_extensions/src_py3/typing_extensions.py
        'NewType': HintSignNewType,

        # ..................{ PEP (484|604)                  }..................
        # Python >= 3.14 implements both PEP 484-compliant old-school unions
        # (e.g., "typing.Union[int, float]") *AND* PEP 604-compliant new-school
        # unions (e.g., "int | float") as instances of the low-level C-based
        # "typing.Union" type. Doing so unifies the syntactic treatment of
        # unions, mildly simplifying union detection: e.g.,
        #     >>> from typing import Optional, Union
        #
        #     >>> type(int | None)
        #     <class 'typing.Union'>  # <-- *GOOD*
        #     >>> type(Union[int, None])
        #     <class 'typing.Union'>  # <-- *GOOD*
        #     >>> type(Optional[int])
        #     <class 'typing.Union'>  # <-- *GOOD*
        #
        #     >>> int | None == Optional[int] == Union[int, None]
        #     True  # <-- woah. CPython mad lads finally did it, huh?
        'Union': HintSignUnion,
    },
})
'''
**Type hint type trie** (i.e., dictionary-of-dictionaries tree data structure
mapping from the fully-qualified names of packages and modules to a nested
dictionary mapping from the unqualified basenames of the types of all
PEP-compliant type hints residing in those packages and modules that are
uniquely identifiable by those types to their identifying signs).
'''

# ....................{ MAPPINGS ~ generics                }....................
# The majority of this dictionary is initialized with automated inspection
# iterating over the "_TYPING_ATTR_HINT_BASENAME_TO_SIGN" dictionary local
# defined by the _init() function below. The *ONLY* key-value pairs explicitly
# defined here are those *NOT* amenable to such inspection.
HINT_MODULE_NAME_TO_HINT_BASE_EXTRINSIC_BASENAME_TO_SIGN = (
    _init_hint_sign_trie({}))
'''
**Extrinsic pseudo-superclass trie** (i.e., dictionary-of-dictionaries tree data
structure mapping from the fully-qualified names of packages and modules to
nested dictionaries mapping from the unqualified basenames of all PEP-compliant
objects residing in those packages and modules such that those objects are valid
extrinsic pseudo-superclasses of :pep:`484`- or :pep:`585`-compliant generics
extrinsically identifiable by signs to those signs).
'''

# ....................{ PRIVATE ~ globals                  }....................
# Note that the builtin "range" class:
# * Initializer "range.__init__(start, stop)" is effectively instantiated as
#   [start, stop) -- that is to say, such that:
#   * The initial "start" integer is *INCLUSIVE* (i.e., the instantiated range
#     includes this integer).
#   * The final "stop" integer is *EXCLUSIVE* (i.e., the instantiated range
#     excludes this integer).
# * Publicizes these integers as the instance variables "start" and "stop" such
#   that these invariants are guaranteed:
#       >>> range(min, max).start == min
#       True
#       >>> range(min, max).stop == max
#       True

_ARGS_LEN_0 = range(0, 1)  # == [0, 1) == [0, 0]
'''
**Zero-argument length range** (i.e., :class:`range` instance effectively
equivalent to the integer ``0``, describing type hint factories subscriptable by
*no* child type hints).
'''


_ARGS_LEN_1 = range(1, 2)  # == [1, 2) == [1, 1]
'''
**One-argument length range** (i.e., :class:`range` instance effectively
equivalent to the integer ``1``, describing type hint factories subscriptable by
exactly one child type hint).
'''


_ARGS_LEN_2 = range(2, 3)  # == [2, 3) == [2, 2]
'''
**Two-argument length range** (i.e., :class:`range` instance effectively
equivalent to the integer ``2``, describing type hint factories subscriptable by
exactly two child type hints).
'''


_ARGS_LEN_3 = range(3, 4)  # == [3, 4) == [3, 3]
'''
**Three-argument length range** (i.e., :class:`range` instance effectively
equivalent to the integer ``3``, describing type hint factories subscriptable by
exactly three child type hints).
'''


_ARGS_LEN_1_OR_2 = range(1, 3)  # == [1, 3) == [1, 2]
'''
**One- or two-argument length range** (i.e., :class:`range` instance effectively
equivalent to the integer range ``[1, 2]``, describing type hint factories
subscriptable by either one or two child type hints).
'''

# ....................{ SIGNS ~ origin : args              }....................
# Fully initialized by the _init() function below.
HINT_SIGN_ORIGIN_ISINSTANCEABLE_TO_ARGS_LEN_RANGE: Dict[HintSign, range] = {
    # Type hint factories subscriptable by exactly one child type hint.
    HintSignAbstractSet: _ARGS_LEN_1,
    HintSignAsyncIterable: _ARGS_LEN_1,
    HintSignAsyncIterator: _ARGS_LEN_1,
    HintSignAwaitable: _ARGS_LEN_1,
    # HintSignByteString: _ARGS_LEN_1,
    HintSignCollection: _ARGS_LEN_1,
    HintSignContainer: _ARGS_LEN_1,
    HintSignCounter: _ARGS_LEN_1,
    HintSignDeque: _ARGS_LEN_1,
    HintSignFrozenSet: _ARGS_LEN_1,
    HintSignIterable: _ARGS_LEN_1,
    HintSignIterator: _ARGS_LEN_1,
    HintSignKeysView: _ARGS_LEN_1,
    HintSignList: _ARGS_LEN_1,
    HintSignMatch: _ARGS_LEN_1,
    HintSignMappingView: _ARGS_LEN_1,
    HintSignMutableSequence: _ARGS_LEN_1,
    HintSignMutableSet: _ARGS_LEN_1,
    HintSignPattern: _ARGS_LEN_1,
    HintSignReversible: _ARGS_LEN_1,
    HintSignSequence: _ARGS_LEN_1,
    HintSignSet: _ARGS_LEN_1,
    HintSignType: _ARGS_LEN_1,
    HintSignValuesView: _ARGS_LEN_1,

    # Type hint factories subscriptable by exactly two child type hints.
    HintSignAsyncGenerator: _ARGS_LEN_2,
    HintSignChainMap: _ARGS_LEN_2,
    HintSignDefaultDict: _ARGS_LEN_2,
    HintSignDict: _ARGS_LEN_2,
    HintSignItemsView: _ARGS_LEN_2,
    HintSignMapping: _ARGS_LEN_2,
    HintSignMutableMapping: _ARGS_LEN_2,
    HintSignOrderedDict: _ARGS_LEN_2,
    HintSignTuple: _ARGS_LEN_2,

    # Type hint factories subscriptable by exactly three child type hints.
    HintSignCoroutine: _ARGS_LEN_3,
    HintSignGenerator: _ARGS_LEN_3,
}
'''
Dictionary mapping from each sign uniquely identifying a PEP-compliant type hint
factory originating from an **isinstanceable origin type** (i.e., isinstanceable
class such that *all* objects satisfying type hints created by subscripting this
factory are instances of this class) to this factory's **argument length range**
(i.e., :class:`range` instance describing the minimum and maximum number of
child type hints that may subscript this factory).
'''

# ....................{ PRIVATE ~ main                     }....................
def _init() -> None:
    '''
    Initialize this submodule.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer initialization-specific imports.
    from beartype._util.py.utilpyversion import (
        IS_PYTHON_AT_MOST_3_13,
        IS_PYTHON_AT_LEAST_3_13,
        IS_PYTHON_3_11,
    )

    # ..................{ LOCALS                             }..................
    # Dictionary mapping from the unqualified names of all callables and classes
    # defined by typing modules that are themselves valid PEP-compliant type
    # hints to their corresponding signs.
    _TYPING_ATTR_HINT_BASENAME_TO_SIGN = {
        # ..................{ PEP 484                        }..................
        # Unsubscripted "typing.Generic" superclass, which imposes no
        # constraints and is also semantically synonymous with the "object"
        # superclass. Since PEP 484 stipulates that *ANY* unsubscripted
        # subscriptable PEP-compliant type hint factories semantically expand to
        # those factories subscripted by an implicit "Any" argument, "Generic"
        # semantically expands to the implicit "Generic[Any]" singleton.
        'Generic': HintSignPep484585GenericUnsubbed,

        # ..................{ PEP 544                        }..................
        # Unsubscripted "typing.Protocol" superclass. For unknown and presumably
        # uninteresting reasons, *ALL* possible objects satisfy this superclass.
        # Ergo, this superclass is synonymous with the "object" root superclass:
        #     >>> from typing import Protocol
        #     >>> isinstance(object(), Protocol)
        #     True
        #     >>> isinstance('wtfbro', Protocol)
        #     True
        #     >>> isinstance(0x696969, Protocol)
        #     True
        'Protocol': HintSignProtocol,
    }

    # Dictionary mapping from the unqualified names of all callables and classes
    # defined by typing modules that are themselves valid extrinsic
    # pseudo-superclasses to their corresponding signs.
    _TYPING_ATTR_HINT_BASE_EXTRINSIC_BASENAME_TO_SIGN = {
        # ..................{ PEP 484                        }..................
        # PEP 484-compliant "typing.NamedTuple" superclass, whose metaclass
        # permits subclasses to also subclass the PEP 484-compliant
        # "typing.Generic" superclass under Python >= 3.11. The resulting
        # user-defined subclasses are referred to as "generic named tuples,"
        # which convey extrinsic type-checking courtesy being named tuples.
        'NamedTuple': HintSignNamedTuple,

        # ..................{ PEP 589                        }..................
        # PEP 589-compliant "typing.TypedDict" superclass, whose metaclass
        # permits subclasses to also subclass the PEP 484-compliant
        # "typing.Generic" superclass under Python >= 3.11. The resulting
        # user-defined subclasses are referred to as "generic typed
        # dictionaries," which convey extrinsic type-checking courtesy being
        # typed dictionaries.
        'TypedDict': HintSignTypedDict,
    }

    # Dictionary mapping from the unqualified names of all classes defined by
    # typing modules used to instantiate PEP-compliant type hints to their
    # corresponding signs.
    _TYPING_ATTR_TYPE_BASENAME_TO_SIGN = {
        # ....................{ PEP 484                    }....................
        # All PEP 484-compliant type variables are necessarily instances of the
        # same class.
        'TypeVar': HintSignTypeVar,

        #FIXME: "Generic" is ignorable when unsubscripted. Excise this up!
        # The unsubscripted PEP 484-compliant "Generic" superclass is
        # explicitly equivalent under PEP 484 to the "Generic[Any]"
        # subscription and thus slightly conveys meaningful semantics.
        # 'Generic': HintSignPep484585GenericUnsubbed,

        # ....................{ PEP 612                    }....................
        # PEP 612-compliant "typing.ParamSpec" type hints as merely instances of
        # that low-level C-based type.
        'ParamSpec': HintSignParamSpec,

        # PEP 612-compliant "*args: P.args" type hints as merely instances of
        # the low-level C-based "typing.ParamSpecArgs" type.
        'ParamSpecArgs': HintSignParamSpecArgs,

        # PEP 612-compliant "**kwargs: P.kwargs" type hints as merely instances
        # of the low-level C-based "typing.ParamSpecKwargs" type.
        'ParamSpecKwargs': HintSignParamSpecKwargs,

        # ....................{ PEP 646                    }....................
        # All PEP 646-compliant type variable tuples are necessarily instances
        # of the same class.
        'TypeVarTuple': HintSignTypeVarTuple,

        # ....................{ PEP 695                    }....................
        # PEP 695-compliant "type" aliases are merely instances of the low-level
        # C-based "typing.TypeAliasType" type.
        'TypeAliasType': HintSignPep695TypeAliasUnsubscripted,
    }

    # ..................{ INIT ~ versions                    }..................
    # If the active Python interpreter targets Python <= 3.13...
    if IS_PYTHON_AT_MOST_3_13:
        # Map both the "typing.ForwardRef" and "typing_extensions.ForwardRef"
        # types (the latter of which is simply an alias of the former) to the
        # sign uniquely identifying forward references. All PEP 484-compliant
        # forward references are necessarily instances of this type.
        #
        # Under Python >= 3.14, both of these types are simply aliases of
        # "annotationlib.ForwardRef" type -- which is now the canonical
        # implementation of this type but which resides outside a typing module.
        _TYPING_ATTR_TYPE_BASENAME_TO_SIGN['ForwardRef'] = HintSignForwardRef
    # Else, the active Python interpreter targets Python >= 3.14.

    # If the active Python interpreter targets Python >= 3.13...
    if IS_PYTHON_AT_LEAST_3_13:
        # Add all signs uniquely identifying one- or two-argument type hint
        # factories under Python >= 3.13, which generalized various one-argument
        # type hint factories to accept an additional optional child type hint
        # via "PEP 696 – Type Defaults for Type Parameters".
        HINT_SIGN_ORIGIN_ISINSTANCEABLE_TO_ARGS_LEN_RANGE.update({
            HintSignAsyncContextManager: _ARGS_LEN_1_OR_2,
            HintSignContextManager: _ARGS_LEN_1_OR_2,
        })
    # Else, the active Python interpreter targets Python <= 3.12. In this
    # case...
    else:
        # Add all signs uniquely identifying two-argument type hint factories
        # under Python <= 3.12.
        HINT_SIGN_ORIGIN_ISINSTANCEABLE_TO_ARGS_LEN_RANGE.update({
            HintSignAsyncContextManager: _ARGS_LEN_1,
            HintSignContextManager: _ARGS_LEN_1,
        })
    # print(f'HINT_SIGN_ORIGIN_ISINSTANCEABLE_TO_ARGS_LEN_RANGE: {HINT_SIGN_ORIGIN_ISINSTANCEABLE_TO_ARGS_LEN_RANGE}')

    # ..................{ INIT ~ modules                     }..................
    # For the fully-qualified name of each quasi-standard typing module...
    for typing_module_name in TYPING_MODULE_NAMES:
        # For the unqualified basename of each type hint that is itself a class
        # or callable identifiable by a sign to that sign, map from the
        # fully-qualified name of that type in this module to this sign.
        for hint_basename, hint_sign in (
            _TYPING_ATTR_HINT_BASENAME_TO_SIGN.items()):
            # print(f'[datahintrepr] Mapping hint "{typing_module_name}.{typing_attr_basename}" -> {hint_sign}')
            HINT_MODULE_NAME_TO_HINT_BASENAME_TO_SIGN[
                typing_module_name][hint_basename] = hint_sign

        # For the unqualified basename of each extrinsic pseudo-superclass of a
        # PEP 484- or 585-compliant generic identifiable by a sign to that sign,
        # map from the fully-qualified name of that type in this module to this
        # sign.
        for hint_basename, hint_sign in (
            _TYPING_ATTR_HINT_BASE_EXTRINSIC_BASENAME_TO_SIGN.items()):
            # print(f'[datahintrepr] Mapping hint "{typing_module_name}.{typing_attr_basename}" -> {hint_sign}')
            HINT_MODULE_NAME_TO_HINT_BASE_EXTRINSIC_BASENAME_TO_SIGN[
                typing_module_name][hint_basename] = hint_sign

        # For the unqualified basename of each type of each type hint
        # identifiable by a sign to that sign, map from the fully-qualified name
        # of that type in this module to this sign.
        for type_basename, hint_sign in (
            _TYPING_ATTR_TYPE_BASENAME_TO_SIGN.items()):
            # print(f'[datahintrepr] Mapping type "{typing_module_name}.{typing_attr_basename}" -> {hint_sign}')
            HINT_MODULE_NAME_TO_TYPE_BASENAME_TO_SIGN[
                typing_module_name][type_basename] = hint_sign

        # If the active Python interpreter targets Python 3.11, identify PEP
        # 646- and 692-compliant hints that are instances of the private
        # "typing._UnpackGenericAlias" as "Unpack[...]" hints.
        #
        # Note that this fragile violation of privacy encapsulation is *ONLY*
        # needed under Python 3.11, where the machine-readable representation of
        # unpacked type variable tuples is ambiguously idiosyncratic and thus
        # *NOT* a reasonable heuristic for detecting such unpacking: e.g.,
        #     $ python3.11
        #     >>> from typing import TypeVarTuple
        #     >>> Ts = TypeVarTuple('Ts')
        #     >>> list_of_Ts = [*Ts]
        #     >>> repr(list_of_Ts[0])
        #     *Ts    # <-- ambiguous and thus a significant issue
        #
        #     $ python3.12
        #     >>> from typing import TypeVarTuple
        #     >>> Ts = TypeVarTuple('Ts')
        #     >>> list_of_Ts = [*Ts]
        #     >>> repr(list_of_Ts[0])
        #     typing.Unpack[Ts]    # <-- unambiguous and thus a non-issue
        if IS_PYTHON_3_11:
            HINT_MODULE_NAME_TO_TYPE_BASENAME_TO_SIGN[
                typing_module_name]['_UnpackGenericAlias'] = HintSignUnpack
        # Else, the active Python interpreter does *NOT* target Python 3.11.

    # ..................{ DEBUGGING                          }..................
    # Uncomment as needed to display the contents of these objects.

    # from pprint import pformat
    # print(f'HINT_MODULE_NAME_TO_TYPE_BASENAME_TO_SIGN: {pformat(HINT_MODULE_NAME_TO_TYPE_BASENAME_TO_SIGN)}')

# Initialize this submodule.
_init()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **integer getteres** (i.e., low-level callables creating and
returning various integers of general-purpose interest).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from random import getrandbits

# ....................{ GETTERS                            }....................
#FIXME: Unit test us up, please.
def get_integer_pseudorandom_signed_32bit() -> int:
    '''
    **Pseudorandom signed 32-bit integer** (i.e., integer in the inclusive range
    ``[-2,147,483,648, 2,147,483,647]`` efficiently generated with the standard
    C-based Mersenne twister :func:`.getrandbits`).

    This integer is sufficiently large to randomly index arbitrary sequences,
    which is exactly the approach leveraged by type-checking wrapper functions
    generated by the :func:`beartype.beartype` decorator.

    Examples
    --------
    .. code-block:: pycon

       # Randomly index an arbitrary sequence efficiently.
       >>> from beartype._util.kind.integer.utilintget import (
       ...     get_integer_pseudorandom_signed_32bit)
       >>> fun_list = ['The', 'funnest', 'list', 'you', 'saw', 'today!']
       >>> fun_list[get_integer_pseudorandom_signed_32bit() % len(fun_list)]
       'funnest'
    '''

    # Tales of One-liners.
    return getrandbits(32)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :mod:`functools` utility unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.api.standard.utilfunctools` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ tester                     }....................
def test_is_func_functools_lru_cache() -> None:
    '''
    Test the
    :func:`beartype._util.api.standard.utilfunctools.is_func_functools_lru_cache`
    tester.
    '''

    # Defer test-specific imports.
    from beartype._util.api.standard.utilfunctools import is_func_functools_lru_cache
    from beartype_test.a00_unit.data.data_type import (
        function_lru_cached,
        function,
    )

    # Assert this tester accepts a @functools.lru_cache-memoized callable.
    assert is_func_functools_lru_cache(function_lru_cached) is True

    # Assert this tester rejects a callable *NOT* decorated by that decorator.
    assert is_func_functools_lru_cache(function) is False

    # Assert this tester rejects a non-callable *WITHOUT* raising exceptions.
    assert is_func_functools_lru_cache(
        'With weeping flowers, or votive cypress wreath,') is False


def test_is_func_functools_partial() -> None:
    '''
    Test the
    :func:`beartype._util.api.standard.utilfunctools.is_func_functools_partial`
    tester.
    '''

    # Defer test-specific imports.
    from beartype._util.api.standard.utilfunctools import is_func_functools_partial
    from beartype_test.a00_unit.data.data_type import (
        function,
        function_partial,
    )

    # Assert this tester accepts a "functools.partial"-wrapped callable.
    assert is_func_functools_partial(function_partial) is True

    # Assert this tester rejects a callable *NOT* decorated by that decorator.
    assert is_func_functools_partial(function) is False

    # Assert this tester rejects a non-callable *WITHOUT* raising exceptions.
    assert is_func_functools_partial(
        'For well he knew that mighty Shadow loves') is False

# ....................{ TESTS ~ getter                     }....................
def test_get_func_functools_partial_args() -> None:
    '''
    Test the
    :func:`beartype._util.api.standard.utilfunctools.get_func_functools_partial_args`
    getter.
    '''

    # Defer test-specific imports.
    from beartype._util.api.standard.utilfunctools import (
        get_func_functools_partial_args)
    from beartype_test.a00_unit.data.data_type import (
        FUNCTION_PARTIALIZED_ARG_VALUE,
        FUNCTION_PARTIALIZED_KWARG_NAME,
        FUNCTION_PARTIALIZED_KWARG_VALUE,
        builtin_partial,
        function_partial,
    )

    # Assert this unwrapper returns the expected 2-tuple providing the
    # positional and keyword parameters passed to the C-based callable wrapped
    # by the "functools.partial" class.
    assert get_func_functools_partial_args(builtin_partial) == ((2,), {})

    # Assert this unwrapper returns the expected 2-tuple providing the
    # positional and keyword parameters passed to the pure-Python callable
    # wrapped by the "functools.partial" class.
    assert get_func_functools_partial_args(function_partial) == (
        (FUNCTION_PARTIALIZED_ARG_VALUE,),
        {FUNCTION_PARTIALIZED_KWARG_NAME: FUNCTION_PARTIALIZED_KWARG_VALUE},
    )

# ....................{ TESTS ~ tester                     }....................
def test_unwrap_func_functools_partial_once() -> None:
    '''
    Test the
    :func:`beartype._util.api.standard.utilfunctools.unwrap_func_functools_partial_once`
    unwrapper.
    '''

    # Defer test-specific imports.
    from beartype._util.api.standard.utilfunctools import (
        unwrap_func_functools_partial_once)
    from beartype_test.a00_unit.data.data_type import (
        builtin_partial,
        function_partialized,
        function_partial,
    )

    # Assert this unwrapper returns the expected C-based callable wrapped by
    # the "functools.partial" class.
    assert unwrap_func_functools_partial_once(builtin_partial) is divmod

    # Assert this unwrapper returns the expected pure-Python callable wrapped by
    # the "functools.partial" class.
    assert unwrap_func_functools_partial_once(function_partial) is (
        function_partialized)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484` and :pep:`585` **forward reference type hint utility
unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep484585.pep484585ref` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ getter                     }....................
def test_get_hint_pep484585_ref_names() -> None:
    '''
    Test
    :func:`beartype._util.hint.pep.proposal.pep484585.pep484585ref.get_hint_pep484585_ref_names`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintForwardRefException
    from beartype.typing import ForwardRef
    from beartype._util.hint.pep.proposal.pep484585.pep484585ref import (
        get_hint_pep484585_ref_names)
    from pytest import raises

    # ....................{ LOCALS                         }....................
    # Arbitrary absolute forward references defined as strings.
    ALAS_ALAS = 'He_overleaps.the.bounds'
    BREATH_AND = 'Were.limbs_and.breath_and'

    # Arbitrary absolute forward reference defined as a non-string.
    BEING_INTERTWINED = ForwardRef(BREATH_AND)

    # Arbitrary fully-qualified names of a hypothetical module.
    OF_DIM_SLEEP = 'In_the.wide.pathless_desert'

    # Arbitrary relative forward reference defined as a string.
    THUS_TREACHEROUSLY = 'Lost_lost_for_ever_lost'

    # ....................{ PASS                           }....................
    # Assert this getter preserves the passed absolute forward reference as is
    # when defined as a string, regardless of whether the second passed object
    # defines the "__module__" dunder attribute.
    assert get_hint_pep484585_ref_names(ALAS_ALAS) == (None, ALAS_ALAS)

    # Assert this getter preserves the passed absolute forward reference as is
    # when defined as a non-string, regardless of whether the second passed
    # object defines the "__module__" dunder attribute.
    assert get_hint_pep484585_ref_names(BEING_INTERTWINED) == (None, BREATH_AND)

    # ....................{ PASS ~ module name             }....................
    # Since the active Python interpreter targets >= Python 3.10, then this
    # typing.ForwardRef.__init__() method accepts an additional optional
    # "module: Optional[str] = None" parameter preserving the fully-qualified
    # name of the module to which the passed unqualified basename is relative.
    # Assert that this getter successfully introspects that module name.

    # Arbitrary relative forward reference defined as a non-string relative to
    # the fully-qualified names of a hypothetical module.
    BEAUTIFUL_SHAPE = ForwardRef(THUS_TREACHEROUSLY, module=OF_DIM_SLEEP)

    # Arbitrary absolute forward reference defined as a non-string relative to
    # the fully-qualified names of a hypothetical module.
    #
    # Note that this exercises an edge case. Technically, passing both an
    # absolute forward reference *AND* a module names is a non-sequitur.
    # Pragmatically, the ForwardRef.__init__() method blindly permits callers to
    # do just that. Ergo, @beartype *MUST* guard against this.
    DARK_GATE_OF_DEATH = ForwardRef(ALAS_ALAS, module=OF_DIM_SLEEP)

    # Assert this getter canonicalizes the passed relative forward reference
    # against the fully-qualified names of the module with which this reference
    # was instantiated when defined as a non-string.
    assert get_hint_pep484585_ref_names(BEAUTIFUL_SHAPE) == (
        OF_DIM_SLEEP, THUS_TREACHEROUSLY)

    # Assert this getter preserves the passed absolute forward reference as is
    # *WITHOUT* canonicalizing this reference against the fully-qualified names
    # of the module with which this reference was instantiated when defined as a
    # non-string.
    assert get_hint_pep484585_ref_names(DARK_GATE_OF_DEATH) == (
        OF_DIM_SLEEP, ALAS_ALAS)

    # ....................{ FAIL                           }....................
    # Assert this getter raises the expected exception when passed a type hint
    # that is *NOT* a valid forward reference.
    with raises(BeartypeDecorHintForwardRefException):
        get_hint_pep484585_ref_names(
            b'Beyond the realms of dream that fleeting shade;')


def test_get_hint_pep484585_ref_names_relative_to() -> None:
    '''
    Test
    :func:`beartype._util.hint.pep.proposal.pep484585.pep484585ref.get_hint_pep484585_ref_names_relative_to`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintForwardRefException
    from beartype.typing import ForwardRef
    from beartype._util.hint.pep.proposal.pep484585.pep484585ref import (
        get_hint_pep484585_ref_names_relative_to)
    from beartype_test.a00_unit.data.data_type import (
        ClassModuleNameFake,
        ClassModuleNameNone,
        function_module_name_fake,
        function_module_name_none,
    )
    from pytest import raises

    # ....................{ LOCALS                         }....................
    # Fully-qualified name of the current module defining this unit test.
    THIS_MODULE_NAME = __name__

    # Arbitrary absolute forward references defined as strings.
    ALAS_ALAS = 'He_overleaps.the.bounds'
    THE_MASK_OF_ANARCHY = 'as_I.lay_asleep.in_Italy'
    WITH_GREAT_POWER_IT = 'And_with.great_power.it_forth.led_me'

    # Arbitrary absolute forward reference defined as a non-string.
    WITH_GREAT_POWER = ForwardRef(WITH_GREAT_POWER_IT)

    # Arbitrary fully-qualified names of hypothetical modules.
    CASTLEREAGH = 'He_had.a_mask.like_Castlereagh'
    OF_DIM_SLEEP = 'In_the.wide.pathless_desert'

    # Arbitrary relative forward references defined as strings.
    VERY_SMOOTH = 'Very_smooth_he_looked_yet_grim'
    THUS_TREACHEROUSLY = 'Lost_lost_for_ever_lost'

    # ....................{ CALLABLES                      }....................
    def and_pendent_mountains():
        '''
        Arbitrary function whose ``__module__`` dunder attribute is preserved as
        the fully-qualified name of this currently importable module.
        '''

        pass

    # ....................{ CLASSES                        }....................
    class OfRainbowClouds(object):
        '''
        Arbitrary class whose ``__module__`` dunder attribute is preserved as
        the fully-qualified name of this currently importable module.
        '''

        pass

    # ....................{ PASS                           }....................
    # Assert that this getter preserves an absolute forward reference in string
    # format as is.
    assert get_hint_pep484585_ref_names_relative_to(THE_MASK_OF_ANARCHY) == (
        None, THE_MASK_OF_ANARCHY)

    # Assert that this getter preserves an absolute forward reference in
    # "typing.ForwardRef" format as is.
    assert get_hint_pep484585_ref_names_relative_to(WITH_GREAT_POWER) == (
        None, WITH_GREAT_POWER_IT)

    # Assert that this getter canonicalizes a relative forward reference against
    # the "__module__" dunder attribute of a function to the expected string.
    assert get_hint_pep484585_ref_names_relative_to(
        hint=VERY_SMOOTH, func=and_pendent_mountains) == (
        THIS_MODULE_NAME, VERY_SMOOTH)

    # Assert that this getter canonicalizes a relative forward reference against
    # the "__module__" dunder attribute of a type stack to the expected string.
    assert get_hint_pep484585_ref_names_relative_to(
        hint=VERY_SMOOTH,
        cls_stack=[OfRainbowClouds],
        func=and_pendent_mountains,
    ) == (THIS_MODULE_NAME, VERY_SMOOTH)

    # Assert that this getter preserves the passed relative forward reference as
    # is when this reference is the name of a builtin type, even when passed
    # neither a class *NOR* callable.
    assert get_hint_pep484585_ref_names_relative_to('int') == (
        'builtins', 'int')

    # Assert that this getter preserves the passed relative forward reference as
    # is when this reference is the name of a builtin type, even when passed
    # a class and callable both defining the "__module__" dunder attribute to be
    # the fully-qualified names of imaginary and thus unimportable modules that
    # do *NOT* physically exist.
    assert get_hint_pep484585_ref_names_relative_to(
        hint='int',
        cls_stack=(ClassModuleNameFake,),
        func=function_module_name_fake,
    ) == ('builtins', 'int')

    # ....................{ PASS ~ module name             }....................
    # Since the active Python interpreter targets >= Python 3.10, then this
    # typing.ForwardRef.__init__() method accepts an additional optional
    # "module: Optional[str] = None" parameter preserving the fully-qualified
    # name of the module to which the passed unqualified basename is relative.
    # Assert that this getter successfully introspects that module name.

    # Arbitrary relative forward reference defined as a non-string relative to
    # the fully-qualified name of a hypothetical module.
    BEAUTIFUL_SHAPE = ForwardRef(THUS_TREACHEROUSLY, module=OF_DIM_SLEEP)

    # Arbitrary absolute forward reference defined as a non-string relative to
    # the fully-qualified name of a hypothetical module.
    #
    # Note that this exercises an edge case. Technically, passing both an
    # absolute forward reference *AND* a module name is a non-sequitur.
    # Pragmatically, the ForwardRef.__init__() method blindly permits callers to
    # do just that. Ergo, @beartype *MUST* guard against this.
    DARK_GATE_OF_DEATH = ForwardRef(ALAS_ALAS, module=OF_DIM_SLEEP)

    # Assert this getter canonicalizes the passed relative forward reference
    # against the fully-qualified name of the module with which this reference
    # was instantiated when defined as a "typing.ForwardRef".
    assert get_hint_pep484585_ref_names_relative_to(BEAUTIFUL_SHAPE) == (
        OF_DIM_SLEEP, THUS_TREACHEROUSLY)

    # Assert this getter preserves the passed absolute forward reference as is
    # *WITHOUT* canonicalizing this reference against the fully-qualified name
    # of the module with which this reference was instantiated when defined as a
    # "typing.ForwardRef".
    assert get_hint_pep484585_ref_names_relative_to(DARK_GATE_OF_DEATH) == (
        OF_DIM_SLEEP, ALAS_ALAS)

    # ....................{ FAIL                           }....................
    # Assert that this getter raises the expected exception when passed a
    # relative forward reference and neither a class *NOR* callable.
    with raises(BeartypeDecorHintForwardRefException):
        get_hint_pep484585_ref_names_relative_to(VERY_SMOOTH)

    # Assert that this getter raises the expected exception when passed a
    # relative forward reference and a class and callable both defining the
    # "__module__" dunder attribute to be the fully-qualified names of imaginary
    # and thus unimportable modules that do *NOT* physically exist.
    with raises(BeartypeDecorHintForwardRefException):
        get_hint_pep484585_ref_names_relative_to(
            hint=VERY_SMOOTH,
            cls_stack=(ClassModuleNameFake,),
            func=function_module_name_fake,
        )

    # Assert that this getter raises the expected exception when passed a
    # relative forward reference and a class and callable both defining the
    # "__module__" dunder attribute to be "None".
    with raises(BeartypeDecorHintForwardRefException):
        get_hint_pep484585_ref_names_relative_to(
            hint=VERY_SMOOTH,
            cls_stack=(ClassModuleNameNone,),
            func=function_module_name_none,
        )

    # Assert that this getter raises the expected exception when passed a
    # relative forward reference and *ONLY* a callable defining the
    # "__module__" dunder attribute to be the fully-qualified name of an
    # imaginary and thus unimportable module that does *NOT* physically exist.
    with raises(BeartypeDecorHintForwardRefException):
        get_hint_pep484585_ref_names_relative_to(
            hint=VERY_SMOOTH, func=function_module_name_fake)

    # Assert that this getter raises the expected exception when passed a
    # relative forward reference and *ONLY* a callable defining the
    # "__module__" dunder attribute to be "None".
    with raises(BeartypeDecorHintForwardRefException):
        get_hint_pep484585_ref_names_relative_to(
            hint=VERY_SMOOTH, func=function_module_name_none)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`544`-compliant **type alias reducers** (i.e., low-level
callables converting higher-level protocols to lower-level type hints more
readily consumable by :mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep544Exception
from beartype._check.metadata.hint.hintsane import (
    HINT_SANE_IGNORABLE,
    HintOrSane,
)
from beartype._data.typing.datatypingport import Hint
from beartype._util.hint.pep.proposal.pep544 import (
    HINT_PEP484_IO_GENERIC_TO_PEP544_PROTOCOL,
    init_HINT_PEP484_IO_GENERIC_TO_PEP544_PROTOCOL,
    is_hint_pep484_generic_io,
    is_hint_pep544_protocol_supertype,
)
from beartype._util.hint.pep.utilpepget import (
    get_hint_pep_origin_or_none,
    get_hint_pep_typeargs_packed,
)

# ....................{ REDUCERS                           }....................
def reduce_hint_pep544(hint: Hint, exception_prefix: str) -> HintOrSane:
    '''
    Reduce the passed :pep:`544`-compliant **protocol** (i.e., user-defined
    subclass of the :class:`typing.Protocol` abstract base class (ABC)) to the
    ignorable :data:`.HINT_SANE_IGNORABLE` singleton if this protocol is a
    parametrization of this ABC by one or more :pep:`484`-compliant **type
    variables** (i.e., :class:`typing.TypeVar` objects).

    As the name implies, this ABC is generic and thus fails to impose any
    meaningful constraints. Since a type variable in and of itself also fails to
    impose any meaningful constraints, these parametrizations are safely
    ignorable in all possible runtime contexts: e.g.,

    .. code-block:: python

       from typing import Protocol, TypeVar
       T = TypeVar('T')
       def noop(param_hint_ignorable: Protocol[T]) -> T: pass

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as reducers cannot be memoized.

    Parameters
    ----------
    hint : Hint
        Type hint to be reduced.
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    All remaining passed keyword parameters are silently ignored.

    Returns
    -------
    HintOrSane
        Lower-level type hint currently supported by :mod:`beartype`.
    '''

    # If this protocol is an unsubscripted
    # "(typing|typing_extension|beartype.typing).Protocol" superclass, then this
    # protocol is ignorable. Reduce this ignorable protocol to the ignorable
    # singleton.
    #
    # For unknown (but probably) uninteresting reasons, *ALL* possible objects
    # satisfy these protocol superclasses. Ergo, these superclasses and *ALL*
    # parametrizations of these superclasses are synonymous with the root
    # "object" superclass: e.g.,
    #     >>> from typing as Protocol
    #     >>> isinstance(object(), Protocol)
    #     True
    #     >>> isinstance('wtfbro', Protocol)
    #     True
    #     >>> isinstance(0x696969, Protocol)
    #     True
    if is_hint_pep544_protocol_supertype(hint):
        return HINT_SANE_IGNORABLE

    # Type originating this protocol if any *OR* "None" otherwise.
    #
    # Note that we intentionally avoid calling the
    # get_hint_pep_origin_type_isinstanceable_or_none() function here, which has
    # been intentionally designed to exclude PEP-compliant type hints
    # originating from "typing" type origins for stability reasons.
    hint_origin = get_hint_pep_origin_or_none(hint)

    # If...
    if (
        # Some type originates this protocol *AND*...
        hint_origin is not None and
        # This origin type is an unsubscripted
        # "(typing|typing_extension|beartype.typing).Protocol" superclass...
        is_hint_pep544_protocol_supertype(hint_origin)
    ):
        # Then the passed protocol is a
        # "(typing|typing_extension|beartype.typing).Protocol" superclass
        # subscripted by one or more PEP 484-compliant type variables (e.g.,
        # "typing.Protocol[T]"). Since these type variables convey *NO*
        # meaningful semantics in this context, this protocol is ignorable by a
        # similar argument as above.
        #
        # Note that protocol superclasses can *ONLY* be parametrized by type
        # variables.
        return HINT_SANE_IGNORABLE
    # Else, this protocol is unignorable.

    # Preserve this unignorable protocol.
    return hint


def reduce_hint_pep484_generic_io_to_pep544_protocol(
    hint: Hint, exception_prefix: str) -> Hint:
    '''
    :pep:`544`-compliant :mod:`beartype` **IO protocol** (i.e., either
    :class:`._Pep544IO` itself *or* a subclass of that class defined by this
    submodule intentionally designed to be usable at runtime) corresponding to
    the passed :pep:`484`-compliant :mod:`typing` **IO generic base class**
    (i.e., either :class:`typing.IO` itself *or* a subclass of
    :class:`typing.IO` defined by the :mod:`typing` module effectively unusable
    at runtime due to botched implementation details).

    This reducer is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner thanks to caching internally performed by this
    reducer.

    Parameters
    ----------
    hint : Hint
        :pep:`484`-compliant :mod:`typing` IO generic base class to be replaced
        by the corresponding :pep:`544`-compliant :mod:`beartype` IO protocol.
    exception_prefix : str
        Human-readable label prefixing the representation of this object in the
        exception message.

    Returns
    -------
    Protocol
        :pep:`544`-compliant :mod:`beartype` IO protocol corresponding to this
        :pep:`484`-compliant :mod:`typing` IO generic base class.

    Raises
    ------
    BeartypeDecorHintPep544Exception
        If this object is *not* a :pep:`484`-compliant IO generic base class.
    '''

    # If this object is *NOT* a PEP 484-compliant "typing" IO generic,
    # raise an exception.
    if not is_hint_pep484_generic_io(hint):
        raise BeartypeDecorHintPep544Exception(
            f'{exception_prefix}type hint {repr(hint)} not '
            f'PEP 484 IO generic base class '
            f'(i.e., "typing.IO", "typing.BinaryIO", or "typing.TextIO").'
        )
    # Else, this object is *NOT* a PEP 484-compliant "typing" IO generic.
    #
    # If this dictionary has yet to be initialized, this submodule has yet to be
    # initialized. In this case, do so.
    #
    # Note that this initialization is intentionally deferred until required.
    # Why? Because this initialization performs somewhat space- and
    # time-intensive work -- including importation of the "beartype.vale"
    # subpackage, which we strictly prohibit importing from global scope.

    #FIXME: Awkward API. Technically, this is fine. It works. So, that's good.
    #Still, the ideal API would be a "dict" subclass that auto-initializes
    #itself on the first call to the dict.get() method called below -- probably
    #by just trivially overriding the dict.get() method to instead perform the
    #logic currently performed by this
    #init_HINT_PEP484_IO_GENERIC_TO_PEP544_PROTOCOL() initialization method.
    elif not HINT_PEP484_IO_GENERIC_TO_PEP544_PROTOCOL:
        init_HINT_PEP484_IO_GENERIC_TO_PEP544_PROTOCOL()
    # In any case, this dictionary is now initialized.

    # PEP 544-compliant IO protocol implementing this PEP 484-compliant IO
    # generic if any *OR* "None" otherwise.
    pep544_protocol = HINT_PEP484_IO_GENERIC_TO_PEP544_PROTOCOL.get(hint)

    # If *NO* PEP 544-compliant IO protocol implements this generic...
    if pep544_protocol is None:
        # Tuple of zero or more type variables parametrizing this hint.
        hint_typevars = get_hint_pep_typeargs_packed(hint)

        #FIXME: Unit test us up, please.
        # If this hint is unparametrized, raise an exception.
        if not hint_typevars:
            raise BeartypeDecorHintPep544Exception(
                f'{exception_prefix}PEP 484 IO generic base class '
                f'{repr(hint)} invalid (i.e., not subscripted (indexed) by '
                f'either "str", "bytes", "typing.Any", or "typing.AnyStr").'
            )
        # Else, this hint is parametrized and thus defines the "__origin__"
        # dunder attribute whose value is the type originating this hint.

        #FIXME: Attempt to actually handle this type variable, please.
        # Reduce this parametrized hint (e.g., "typing.IO[typing.AnyStr]") to
        # the equivalent unparametrized hint (e.g., "typing.IO"), effectively
        # ignoring the type variable parametrizing this hint.
        hint_unparametrized: type = get_hint_pep_origin_or_none(hint)  # type: ignore[assignment]

        #FIXME: The caching-specific assignment
        #"HINT_PEP484_IO_GENERIC_TO_PEP544_PROTOCOL[hint] = \" should no longer
        #be required (or desired), as this reducer is itself memoized.
        # PEP 544-compliant IO protocol implementing this unparametrized PEP
        # 484-compliant IO generic. For efficiency, we additionally cache this
        # mapping under the original parametrized hint to minimize the cost of
        # similar reductions under subsequent annotations.
        pep544_protocol = \
            HINT_PEP484_IO_GENERIC_TO_PEP544_PROTOCOL[hint] = \
            HINT_PEP484_IO_GENERIC_TO_PEP544_PROTOCOL[hint_unparametrized]
    # Else, some PEP 544-compliant IO protocol implements this generic.

    # Return this protocol.
    return pep544_protocol



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`649`-compliant **data factories** (i.e., low-level callables
creating and returning lists of :pep:`649`-compliant type hints, exercising edge
cases in unit tests requiring these fixtures).
'''

# ....................{ FIXTURES                           }....................
def hints_pep649_meta() -> 'List[HintPepMetadata]':
    '''
    List of :pep:`649`-compliant **type hint metadata** (i.e.,
    :class:`beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata`
    instances describing test-specific :pep:`649`-compliant sample type hints
    with metadata generically leveraged by various PEP-agnostic unit tests).
    '''

    # ..................{ IMPORTS                            }..................
    # Defer fixture-specific imports.
    from beartype.typing import ForwardRef
    from beartype._data.hint.sign.datahintsigns import HintSignForwardRef
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_MOST_3_13
    from beartype_test.a00_unit.data.data_type import Subclass
    from beartype_test.a00_unit.data.hint.util.data_hintmetacls import (
        HintPepMetadata,
        HintPithSatisfiedMetadata,
        HintPithUnsatisfiedMetadata,
    )

    # ....................{ CONSTANTS ~ forwardref         }....................
    # Fully-qualified classname of an arbitrary class guaranteed to be
    # importable.
    _FORWARDREF_CLASSNAME = 'beartype_test.a00_unit.data.data_type.Subclass'

    # Arbitrary class referred to by "FORWARDREF_CLASSNAME".
    _FORWARDREF_CLASSNAME_TYPE = Subclass

    # ..................{ LISTS                              }..................
    # List of all PEP-specific type hint metadata to be returned.
    hints_pep_meta = [
        # ................{ FORWARDREF                         }................
        # Forward references defined below are *ONLY* intended to shallowly
        # exercise support for types of forward references across the codebase;
        # they are *NOT* intended to deeply exercise resolution of forward
        # references to undeclared classes, which requires more finesse.
        #
        # See the "data_hintref" submodule for the latter.

        # PEP 484-compliant forward reference specified as a fully-qualified
        # classname.
        HintPepMetadata(
            hint=_FORWARDREF_CLASSNAME,
            pep_sign=HintSignForwardRef,
            is_type_typing=False,
            piths_meta=(
                # Instance of the class referred to by this reference.
                HintPithSatisfiedMetadata(_FORWARDREF_CLASSNAME_TYPE()),
                # String constant.
                HintPithUnsatisfiedMetadata(
                    'Empirical Ṗath after‐mathematically harvesting agro‐'),
            ),
        ),

        # PEP 649-compliant forward reference encapsulated by a forward
        # reference proxy published by the "annotationlib" module.
        HintPepMetadata(
            hint=ForwardRef(_FORWARDREF_CLASSNAME),
            pep_sign=HintSignForwardRef,
            # The module defining the "ForwardRef" type differs depending on
            # Python version. Specifically:
            # * If the active Python interpreter targets Python >= 3.14, the
            #   canonical definition resides at "annotationlib.ForwardRef".
            # * If the active Python interpreter targets Python <= 3.13, the
            #   canonical definition resides at "typing.ForwardRef".
            is_type_typing=IS_PYTHON_AT_MOST_3_13,
            piths_meta=(
                # Instance of the class referred to by this reference.
                HintPithSatisfiedMetadata(_FORWARDREF_CLASSNAME_TYPE()),
                # String constant.
                HintPithUnsatisfiedMetadata('Silvicultures of'),
            ),
        ),
    ]

    # ..................{ RETURN                             }..................
    # Return this list of all PEP-specific type hint metadata.
    return hints_pep_meta
